from policyengine_us.model_api import *


class ny_inflation_refund_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "New York 2025 inflation refund credits"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606#QQQ"
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        return 0

    def formula_2025(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.ny.tax.income.credits.inflation_refund
        agi = tax_unit("ny_agi", period)
        filing_status = tax_unit("filing_status", period)
        filing_statuses = filing_status.possible_values

        return select(
            [
                filing_status == filing_statuses.SINGLE,
                filing_status == filing_statuses.JOINT,
                filing_status == filing_statuses.HEAD_OF_HOUSEHOLD,
                filing_status == filing_statuses.SEPARATE,
                filing_status == filing_statuses.SURVIVING_SPOUSE,
            ],
            [
                p.single.calc(agi, right=True),
                p.joint.calc(agi, right=True),
                p.head_of_household.calc(agi, right=True),
                p.separate.calc(agi, right=True),
                p.surviving_spouse.calc(agi, right=True),
            ],
        )

    def formula_2026(tax_unit, period, parameters):
        return 0



from policyengine_us.model_api import *


class emp_self_emp_ratio(Variable):
    value_type = float
    entity = Person
    label = "Share of earnings from wages and salaries"
    unit = "/1"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/1402"

    def formula(person, period, parameters):
        employment_income = person("employment_income", period)
        self_employment_income = person("self_employment_income", period)
        earnings = employment_income + self_employment_income
        res = np.ones_like(earnings)
        mask = earnings > 0
        res[mask] = employment_income[mask] / earnings[mask]
        return res



from policyengine_us.model_api import *


class ia_qbi_deduction(Variable):
    value_type = float
    entity = Person
    label = "Iowa deduction that is fraction of federal qualified business income deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2022-01/IA1040%2841-001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2022IA1040%2841001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf"
    )
    defined_for = StateCode.IA

    def formula(person, period, parameters):
        fed_indv_qbid = person("qbid_amount", period)
        tax_unit = person.tax_unit
        fed_unit_qbid = tax_unit("qualified_business_income_deduction", period)
        # avoid divide-by-zero warnings by not using where() function
        reduction_factor = np.ones_like(fed_indv_qbid)
        mask = fed_indv_qbid > 0
        reduction_factor[mask] = fed_unit_qbid[mask] / fed_indv_qbid[mask]
        fed_qbid = fed_indv_qbid * reduction_factor
        p = parameters(period).gov.states.ia.tax.income
        return fed_qbid * p.deductions.qualified_business_income.fraction



from policyengine_us.model_api import *


class net_investment_income(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "net investment income (NII) that is base of the NII Tax (NIIT)"
    reference = "https://www.law.cornell.edu/uscode/text/26/1411"
    unit = USD

    adds = "gov.irs.investment.income.sources"



from policyengine_us.model_api import *


class unreported_payroll_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Unreported payroll taxes from Form 4137 or 8919"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class new_clean_vehicle_credit_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    label = "Eligible for new clean vehicle credit"
    documentation = "Eligible for nonrefundable credit for the purchase of a new clean vehicle"
    unit = USD
    reference = "https://www.law.cornell.edu/uscode/text/26/30D"
    defined_for = "purchased_qualifying_new_clean_vehicle"

    def formula(tax_unit, period, parameters):
        # Capacity limit applies with and without the Inflation Reduction Act.
        capacity = tax_unit("new_clean_vehicle_battery_capacity", period)
        p = parameters(period).gov.irs.credits.clean_vehicle.new.eligibility
        meets_capacity_requirement = capacity >= p.min_kwh
        agi = tax_unit("adjusted_gross_income", period)
        filing_status = tax_unit("filing_status", period)
        meets_income_limit = agi <= p.income_limit[filing_status]
        msrp = tax_unit("new_clean_vehicle_msrp", period)
        classification = tax_unit("new_clean_vehicle_classification", period)
        meets_msrp_limit = msrp <= p.msrp_limit[classification]
        return (
            meets_capacity_requirement & meets_income_limit & meets_msrp_limit
        )



from policyengine_us.model_api import *


class ca_capi(Variable):
    value_type = float
    unit = USD
    entity = SPMUnit
    label = "California CAPI"
    definition_period = YEAR
    defined_for = "ca_capi_eligible"
    reference = "https://www.cdss.ca.gov/Portals/9/CAPI/CAPI_Regulations-Accessible.pdf"

    def formula(spm_unit, period, parameters):
        state_supplement_payment_standard = spm_unit(
            "ca_state_supplement", period
        )
        person = spm_unit.members
        ssi_amount_if_eligible = person("ssi_amount_if_eligible", period)
        eligible_person = person("ca_capi_eligible_person", period)
        total_ssi_amount = spm_unit.sum(
            ssi_amount_if_eligible * eligible_person
        )
        payment_standard = state_supplement_payment_standard + total_ssi_amount
        p = parameters(period).gov.states.ca.cdss.capi.payment_standard_offset
        married = spm_unit("spm_unit_is_married", period)
        offset = where(married, p.couple, p.single)
        total_payment_standard = max_(payment_standard - offset, 0)
        person = spm_unit.members
        countable_income = spm_unit.sum(
            (person("ssi_countable_income", period) * eligible_person)
        )
        return max_(0, total_payment_standard - countable_income)



from policyengine_us.model_api import *


class ssi_unearned_income_deemed_from_ineligible_spouse(Variable):
    value_type = float
    entity = Person
    label = "SSI unearned income (deemed from ineligible spouse)"
    documentation = "This is ignored if total income is under the SSI individual allowance."
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/cfr/text/20/416.1163"

    def formula(person, period, parameters):
        # First, determine unearned income from the ineligible spouse.
        # This is (a) in the law.
        ineligible_spouse = person("is_ssi_ineligible_spouse", period)
        unearned_income = person("ssi_unearned_income", period)
        ineligible_spousal_unearned_income = (
            person.marital_unit.sum(ineligible_spouse * unearned_income)
            - ineligible_spouse * unearned_income
        )

        # Next, (b) determine and subtract allocations for ineligible children.
        child_allocations = add(
            person.tax_unit, period, ["ssi_ineligible_child_allocation"]
        )
        return max_(0, ineligible_spousal_unearned_income - child_allocations)



from policyengine_us.model_api import *


class taxable_pension_income(Variable):
    value_type = float
    entity = Person
    label = "taxable pension income"
    unit = USD
    definition_period = YEAR
    uprating = "calibration.gov.irs.soi.taxable_pension_income"

    adds = ["taxable_public_pension_income", "taxable_private_pension_income"]



from policyengine_us.model_api import *


class employment_income(Variable):
    value_type = float
    entity = Person
    label = "employment income"
    documentation = "Wages and salaries, including tips and commissions."
    unit = USD
    definition_period = YEAR
    adds = [
        "employment_income_before_lsr",
        "employment_income_behavioral_response",
    ]
    reference = "https://www.law.cornell.edu/uscode/text/26/3401#a"
    uprating = "calibration.gov.irs.soi.employment_income"



from policyengine_us.model_api import *


class az_property_tax_credit_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Income for the Arizona property tax the credit"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.AZ

    adds = "gov.states.az.tax.income.credits.property_tax.income_sources"



from policyengine_us.model_api import *


class meets_snap_work_requirements(Variable):
    value_type = bool
    entity = SPMUnit
    label = "SPM Unit is eligible for SNAP benefits via work requirements"
    definition_period = MONTH
    reference = "https://www.fns.usda.gov/snap/work-requirements"

    def formula(spm_unit, period, parameters):
        p = parameters(
            period
        ).gov.usda.snap.work_requirements.abawd.age_threshold
        person = spm_unit.members
        general_work_requirements = person(
            "meets_snap_general_work_requirements", period
        )
        abawd_work_requirements = person(
            "meets_snap_abawd_work_requirements", period
        )
        # If there is no dependent child, then the SPM unit must meet both work requirements.
        age = person("monthly_age", period)
        is_dependent = person("is_tax_unit_dependent", period)
        is_child = age < p.dependent
        no_dependent_child = person.spm_unit.sum(is_dependent & is_child) == 0
        meets_work_requirements_person = where(
            no_dependent_child,
            abawd_work_requirements & general_work_requirements,
            general_work_requirements,
        )

        return spm_unit.sum(~meets_work_requirements_person) == 0



from policyengine_us.model_api import *


class ne_school_readiness_credit_eligible_worker(Variable):
    value_type = bool
    entity = Person
    label = "Eligible worker for the Nebraska school readiness refundable tax credit"
    definition_period = YEAR
    reference = "https://revenue.nebraska.gov/sites/default/files/doc/tax-forms/2024/f_Individual_Income_Tax_Booklet.pdf#page=2"
    defined_for = StateCode.NE
    default_value = False



from policyengine_us.model_api import *


class is_surviving_child_of_disabled_veteran(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    label = "Surviving child of disabled veteran"



from policyengine_us.model_api import *


class ok_federal_eitc_maximum(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maximum federal EITC for the Oklahoma EITC computation"
    definition_period = YEAR
    reference = (
        "https://law.justia.com/codes/oklahoma/title-68/section-68-2357-43/"
    )
    unit = USD
    defined_for = StateCode.OK

    def formula(tax_unit, period, parameters):
        child_count = tax_unit("eitc_child_count", period)
        eitc = parameters(f"2020-01-01").gov.irs.credits.eitc
        return eitc.max.calc(child_count)



from policyengine_us.model_api import *


class tax_unit_income_ami_ratio(Variable):
    value_type = float
    entity = TaxUnit
    label = "Ratio of tax unit income to area median income"
    definition_period = YEAR



from policyengine_us.model_api import *
from policyengine_us.variables.gov.hhs.tax_unit_fpg import fpg


class tax_unit_medicaid_income_level(Variable):
    value_type = float
    entity = TaxUnit
    label = "Medicaid/CHIP-related modified adjusted gross income (MAGI) level"
    unit = "/1"
    documentation = (
        "Medicaid/CHIP-related MAGI as fraction of federal poverty line."
        "Documentation: 'Federal poverty level (FPL)' at the following URL:"
        "URL: https://www.healthcare.gov/glossary/federal-poverty-level-fpl/"
        "**Pregnant Women:**"
        "  * Pregnant women are counted as themselves plus the number of children they are expecting to deliver"
        "    when determining household size for Medicaid eligibility."
        "  * Sources:"
        "      URL: https://www.sos.state.co.us/CCR/GenerateRulePdf.do?ruleVersionId=11618&fileName=10%20CCR%25202505-10%208.100"
        "      URL: https://www.cms.gov/marketplace/technical-assistance-resources/special-populations-pregnant-women.pdf"
    )
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        income = tax_unit("medicaid_magi", period)

        pregnant_count = add(tax_unit, period, ["current_pregnancies"])
        tax_unit_size = tax_unit("tax_unit_size", period)
        state_group = tax_unit.household("state_group_str", period)

        medicaid_fpg = fpg(
            pregnant_count + tax_unit_size, state_group, period, parameters
        )

        return income / medicaid_fpg



from policyengine_us.model_api import *


class ny_drive_clean_vehicle_electric_range(Variable):
    value_type = float
    entity = Household
    label = "New York Drive Clean rebate program all-electric vehicle range"
    unit = "miles"
    definition_period = YEAR
    reference = "https://www.nyserda.ny.gov/-/media/Project/Nyserda/Files/Programs/Drive-Clean-NY/implementation-manual.pdf#page=8"
    defined_for = StateCode.NY



from policyengine_us.model_api import *


class dc_tanf_basic_eligibility_requirements(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Basic eligibility requirements for DC Temporary Assistance for Needy Families (TANF)"
    definition_period = MONTH
    reference = "https://dhs.dc.gov/service/tanf-district-families"
    defined_for = StateCode.DC

    def formula(spm_unit, period, parameters):
        demographic_eligible = (
            add(spm_unit, period, ["dc_tanf_demographic_eligible_person"]) > 0
        )
        income_eligible = spm_unit("dc_tanf_income_eligible", period)
        resources_eligible = spm_unit("dc_tanf_resources_eligible", period)
        return demographic_eligible & income_eligible & resources_eligible



from policyengine_us.model_api import *


class separate_filer_itemizes(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    label = "Separate filer itemizes"
    documentation = "Whether the taxpayer in this tax unit has a spouse who files separately and itemizes deductions."



from policyengine_us.model_api import *


class nyc_cdcc_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for NYC CDCC"
    definition_period = YEAR
    defined_for = "in_nyc"

    def formula(tax_unit, period, parameters):
        # Eligibility is based on having a federal AGI below $30k
        # and being an NYC full-time resident.

        # First get their income.
        income = tax_unit("adjusted_gross_income", period)

        # Then get the CDCC part of the parameter tree.
        p = parameters(period).gov.local.ny.nyc.tax.income.credits.cdcc

        return income <= p.phaseout_end



"""
This module contains a test that iterates through all parameter files in the policyengine_us/parameters directory and asserts that none of them contain the '\t' character.
"""

from policyengine_us.model_api import REPO


def test_parameter_files_do_not_contain_tabs():
    """
    This test iterates through all parameter files in the policyengine_us/parameters directory and asserts that none of them contain the '\t' character.
    """
    errors = []
    for file_name in (REPO / "parameters").glob("**/*.yaml"):
        with file_name.open() as file:
            i = 0
            for line in file:
                i += 1
                if "\t" in line:
                    errors.append(
                        f"\n\n{file_name.relative_to(REPO)} (line {i}):"
                        + "'\\t' character found in line (shown below):\n\n"
                        + line.replace("\t", "[TAB]")
                    )

    num_errors = len(errors)
    assert num_errors == 0, (
        f"\nFound {len(errors)} parameter file{'s' if len(errors) != 1 else ''} with tabs. Details below:"
        + "\n".join(errors)
    )



from policyengine_us.model_api import *
import numpy as np


class is_infant_for_medicaid_fc(Variable):
    value_type = bool
    entity = Person
    label = "Medicaid infant financial criteria"
    definition_period = YEAR

    def formula(person, period, parameters):
        p = parameters(period).gov.hhs.medicaid.eligibility.categories.infant

        income = person("medicaid_income_level", period)
        state_code = person.household(
            "state_code_str", period
        )  # e.g. "CA", "NY"
        income_limit = p.income_limit[state_code]
        # allow exactly at the threshold
        return np.isclose(income, income_limit) | (income <= income_limit)



from policyengine_us.model_api import *


class deep_poverty_gap(Variable):
    value_type = float
    entity = SPMUnit
    label = "deep poverty gap"
    unit = USD
    documentation = (
        "Difference between household income and deep poverty line."
    )
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        income = spm_unit("spm_unit_net_income", period)
        deep_poverty_threshold = spm_unit("deep_poverty_line", period)
        return max_(deep_poverty_threshold - income, 0)



from policyengine_us.model_api import *


class co_sales_tax_refund_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Colorado sales tax refund"
    definition_period = YEAR
    reference = "https://tax.colorado.gov/sites/tax/files/documents/DR_0104_Book_2022.pdf#page=23"
    defined_for = StateCode.CO

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.co.tax.income.credits.sales_tax_refund
        age_head = tax_unit("age_head", period)
        head_age_eligible = age_head >= p.age_threshold
        age_spouse = tax_unit("age_spouse", period)
        spouse_age_eligible = age_spouse >= p.age_threshold
        # Legal code is ambiguous, but the form points to the line corresponding to
        # tax before non-refundable credits.
        income_tax_eligible = (
            tax_unit("co_income_tax_before_non_refundable_credits", period) > 0
        )
        # Filers that had Colorado tax withheld are also eligible.
        # Use employment income as a proxy.
        employment_income_eligible = (
            add(tax_unit, period, ["irs_employment_income"]) > 0
        )
        income_eligible = employment_income_eligible | income_tax_eligible
        age_eligible = head_age_eligible | spouse_age_eligible
        return age_eligible | income_eligible



from policyengine_us.model_api import *


class wv_low_income_earned_income_exclusion(Variable):
    value_type = float
    entity = TaxUnit
    label = "West Virginia low-income earned income exclusion"
    defined_for = "wv_low_income_earned_income_exclusion_eligible"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://code.wvlegislature.gov/11-21-10/"
        "https://tax.wv.gov/Documents/TaxForms/2021/it140.pdf#page=20"
        "https://tax.wv.gov/Documents/PIT/2022/PersonalIncomeTaxFormsAndInstructions.2022.pdf#page=27"
    )

    def formula(tax_unit, period, parameters):
        federal_agi = tax_unit("adjusted_gross_income", period)
        earned_income = tax_unit("tax_unit_earned_income", period)
        filing_status = tax_unit("filing_status", period)

        p = parameters(
            period
        ).gov.states.wv.tax.income.subtractions.low_income_earned_income
        # Lesser of federal AGI, earned income, and exclusion cap.
        income_min = min_(federal_agi, earned_income)
        return min_(income_min, p.amount[filing_status])



from policyengine_us.model_api import *


class az_family_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arizona Family Tax Credit"
    unit = USD
    definition_period = YEAR
    defined_for = "az_family_tax_credit_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.az.tax.income.credits.family_tax_credits.amount
        filing_status = tax_unit("filing_status", period)

        amount = p.per_person * tax_unit("tax_unit_size", period)
        return min_(amount, p.cap[filing_status])



from policyengine_us.model_api import *


class is_infant_for_medicaid_nfc(Variable):
    value_type = bool
    entity = Person
    label = "Medicaid infant non-financial criteria"
    definition_period = YEAR

    def formula(person, period, parameters):
        age = person("age", period)  # person’s age in years
        state = person.household("state_code_str", period)

        p = parameters(
            period
        ).gov.hhs.medicaid.eligibility.categories.infant.age_range

        # Pick the right upper-bound age for this state
        upper_bound = select(
            [state == "CA", state == "MN"],
            [p.in_ca, p.in_mn],  #
            default=p.other,
        )

        return age < upper_bound



from policyengine_us.model_api import *


class co_tanf_grant_standard(Variable):
    value_type = float
    entity = SPMUnit
    label = "Colorado TANF grant standard"
    unit = USD
    definition_period = YEAR
    defined_for = "co_tanf_eligible"

    def formula(spm_unit, period, parameters):
        children = spm_unit("co_tanf_count_children", period)
        adults = spm_unit("spm_unit_count_adults", period)
        count_pregnant_people = add(spm_unit, period, ["is_pregnant"])
        p = parameters(period).gov.states.co.cdhs.tanf.grant_standard
        capped_children = min_(children, 10).astype(int)
        capped_adults = min_(adults, 2).astype(int)
        additional_children = children - capped_children
        base = p.main[capped_adults][capped_children]
        additional_grant_standard = p.additional_child * additional_children
        pregnancy_allowance = p.pregnancy_allowance * count_pregnant_people
        monthly = base + additional_grant_standard + pregnancy_allowance
        return monthly * MONTHS_IN_YEAR



from policyengine_us.model_api import *


class itemized_taxable_income_deductions_reduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Itemized taxable income deductions reduction"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.deductions.itemized.reduction
        if p.applies:
            agi = tax_unit("adjusted_gross_income", period)

            filing_status = tax_unit("filing_status", period)
            agi_threshold = p.agi_threshold[filing_status]
            agi_excess = max_(0, agi - agi_threshold)
            agi_excess_reduction = agi_excess * p.rate.excess_agi
            maximum_deductions = tax_unit(
                "total_itemized_taxable_income_deductions", period
            )
            maximum_deductions_reduction = maximum_deductions * p.rate.base
            if p.amended_structure.applies:
                top_rate_threshold = parameters(
                    period
                ).gov.irs.income.bracket.thresholds["6"][filing_status]
                exemptions = tax_unit("exemptions", period)
                taxable_income = max_(0, agi - exemptions)
                taxable_income_excess = max_(
                    0, taxable_income - top_rate_threshold
                )
                total_itemized_deductions = tax_unit(
                    "total_itemized_taxable_income_deductions", period
                )
                lesser_of_deductions_or_excess = min_(
                    total_itemized_deductions, taxable_income_excess
                )
                return (
                    p.amended_structure.rate * lesser_of_deductions_or_excess
                )
            return min_(agi_excess_reduction, maximum_deductions_reduction)
        return 0



from policyengine_us.model_api import *


class pre_subsidy_childcare_expenses(Variable):
    value_type = float
    entity = Person
    label = "Pre subsidy child care expenses"
    definition_period = YEAR
    unit = USD

    def formula(person, period, parameters):
        # distribute the SPM unit's childcare expenses evenly across
        # children in SPM unit's Tax units
        spm_unit = person.spm_unit
        childcare_expenses = spm_unit(
            "spm_unit_pre_subsidy_childcare_expenses", period
        )
        is_child = person("is_child", period)
        spm_unit_count_children = add(spm_unit, period, ["is_child"])
        # avoid array divide-by-zero warning by not using where() function
        # see the following GitHub issue for more details:
        # https://github.com/PolicyEngine/policyengine-us/issues/2494
        prorated_expenses = np.zeros_like(spm_unit_count_children)
        mask = spm_unit_count_children > 0
        prorated_expenses[mask] = (
            childcare_expenses[mask] / spm_unit_count_children[mask]
        )
        return prorated_expenses * is_child



from policyengine_us.model_api import *


class has_phone_expense(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Has phone costs"
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        return add(spm_unit, period, ["phone_expense"]) > 0



from policyengine_us.model_api import *


class ut_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Utah non-refundable tax credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.UT

    adds = "gov.states.ut.tax.income.credits.non_refundable"



from IPython.display import Markdown, display_markdown
import pandas as pd
import plotly.express as px
import warnings

warnings.filterwarnings("ignore")

from policyengine_core.enums import Enum
from policyengine_core.charts import (
    LIGHTER_BLUE,
    LIGHT_BLUE,
    BLUE,
    DARK_BLUE,
    format_fig,
    display_fig,
)
from policyengine_us.system import system
from policyengine_us import Simulation


def variable_summary(variable_name: str):
    variable = system.variables.get(variable_name)
    metadata_df = pd.DataFrame(
        {
            "Name": [variable.name],
            "Label": [variable.label],
            "Entity": [variable.entity.key],
            "Definition period": [variable.definition_period],
            "Has a formula": [variable.formula is not None],
            "Value type": [variable.value_type.__name__],
            "Unit": [variable.unit],
            "Documentation": [variable.documentation],
            "Reference": [variable.reference],
            "Quantity type": [variable.quantity_type],
        }
    )

    if variable.defined_for is not None:
        metadata_df["Defined for"] = [f"`{variable.defined_for}`"]
    else:
        metadata_df["Defined for"] = ["All entities"]

    if variable.value_type == Enum:
        metadata_df["Possible values"] = [variable.possible_values.__name__]
        metadata_df["Default value"] = [variable.default_value]

    metadata_df = metadata_df.T.rename(columns={0: "Value"})
    metadata_df.index.name = "Metadata"
    metadata_df = metadata_df.fillna("None")

    return metadata_df


def variation_chart(
    variable: str,
    axis: str = "employment_income",
    y_axis_min: float = 0,
    y_axis_max: float = None,
    state: str = "CA",
    in_notebook: bool = False,
    additional_adult_data: dict = None,
    additional_child_data: dict = None,
    child_count_list: list = [0, 1, 2, 3],
):
    dfs = []

    if additional_adult_data is None:
        additional_adult_data = {}

    if additional_child_data is None:
        additional_child_data = {}

    adult_template = dict(
        age=30,
        **additional_adult_data,
    )

    child_template = dict(
        age=10,
        **additional_child_data,
    )

    for adults in [1, 2]:
        for children in child_count_list:
            people = dict()
            for i in range(adults):
                people[f"adult_{i}"] = adult_template
            for i in range(children):
                people[f"child_{i}"] = child_template
            situation = dict(
                people=people,
                household=dict(
                    state_code=state,
                ),
                axes=[[dict(name=axis, min=0, max=100_000, count=101)]],
            )
            sim = Simulation(situation=situation)
            earnings = sim.calculate(axis, map_to="spm_unit")
            values = sim.calculate(variable, map_to="spm_unit")
            dfs += [
                pd.DataFrame(
                    {
                        "Adults": adults,
                        "Children": children,
                        system.variables[axis].label: earnings,
                        system.variables[variable].label: values,
                    }
                )
            ]

    df = pd.concat(dfs)

    fig = px.line(
        df,
        x=system.variables[axis].label,
        y=system.variables[variable].label,
        color="Children",
        facet_col="Adults",
        facet_col_spacing=0.1,
        color_discrete_sequence=[LIGHTER_BLUE, LIGHT_BLUE, BLUE, DARK_BLUE],
    )
    fig.update_xaxes(
        tickformat="$,.0f",
    )
    max_data_value = max(df[system.variables[variable].label])
    # Round up to nearest value in same order of magnitude
    max_data_value = 10 ** (len(str(int(max_data_value))) - 1) * (
        1 + int(max_data_value / 10 ** (len(str(int(max_data_value))) - 1))
    )
    fig.update_yaxes(
        tickformat="$,.0f",
        range=(y_axis_min, y_axis_max or max_data_value),
    )

    fig.layout.annotations[0].text = "Single"
    fig.layout.annotations[1].text = "Couple"

    if in_notebook:
        return format_fig(fig)
    return display_fig(format_fig(fig))



from policyengine_us.model_api import *


class co_state_addback(Variable):
    value_type = float
    entity = TaxUnit
    label = "Colorado state income tax addback"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.colorado.gov/sites/tax/files/documents/DR_104_Book_2021.pdf#page=5"
        "https://tax.colorado.gov/sites/tax/files/documents/DR_104_Book_2022.pdf#page=5"
    )
    defined_for = StateCode.CO

    def formula(tax_unit, period, parameters):
        federal_itemizer = tax_unit("tax_unit_itemizes", period)
        state_inctax = max_(0, tax_unit("state_withheld_income_tax", period))
        property_taxes = add(tax_unit, period, ["real_estate_taxes"])
        # follow worksheet on page 5 of 2021 Book cited above:
        irs_schA_line_5d = state_inctax + property_taxes
        irs_schA_line_5e = tax_unit("salt_deduction", period)
        ws_line_a = where(
            irs_schA_line_5d > irs_schA_line_5e,
            max_(0, irs_schA_line_5e - property_taxes),
            state_inctax,
        )
        p = parameters(period).gov.irs.deductions
        ws_line_b = add(tax_unit, period, p.itemized_deductions)
        ws_line_c = tax_unit("standard_deduction", period)
        ws_line_d = max_(0, ws_line_b - ws_line_c)
        return federal_itemizer * min_(ws_line_a, ws_line_d)



from policyengine_us.model_api import *


class nc_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "North Carolina deductions"
    unit = USD
    definition_period = YEAR
    reference = "https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions"
    defined_for = StateCode.NC

    adds = "gov.states.nc.tax.income.deductions.deductions"



from policyengine_us.model_api import *


class nyc_cdcc_applicable_percentage(Variable):
    value_type = float
    entity = TaxUnit
    label = "NYC CDCC rate"
    unit = USD
    definition_period = YEAR
    defined_for = "in_nyc"

    def formula(tax_unit, period, parameters):
        # Calculate the "applicable percentage" of the portion of the NYS CDCC
        # applied for children under age four.
        # Up to $25k, the percentage is 75%. Above $25k, it phases out to 0% above $30k.
        # See https://codelibrary.amlegal.com/codes/newyorkcity/latest/NYCadmin/0-0-0-13608
        # § 11-1706 Credits against tax (e).

        # First get federal AGI.
        income = tax_unit("adjusted_gross_income", period)

        # Then get the CDCC part of the parameter tree.
        p = parameters(period).gov.local.ny.nyc.tax.income.credits.cdcc

        phase_out_width = p.phaseout_end - p.phaseout_start
        excess = max_(income - p.phaseout_start, 0)
        capped_excess = min_(excess, phase_out_width)
        percent_excess = capped_excess / phase_out_width
        return p.max_rate * (1 - percent_excess)



from policyengine_us.model_api import *


class reduced_price_school_meals(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    label = "reduced price school meals"
    unit = USD
    documentation = "Value of reduced price school meals."

    def formula(spm_unit, period, parameters):
        tier = spm_unit("school_meal_tier", period)
        is_reduced_price = tier == tier.possible_values.REDUCED
        return is_reduced_price * spm_unit("school_meal_net_subsidy", period)



from policyengine_us.model_api import *


class ky_itemized_deductions_unit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Kentucky itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://revenue.ky.gov/Forms/Form%20740%20Schedule%20A%202022.pdf"
        "https://law.justia.com/codes/kentucky/2022/chapter-141/section-141-019/"  # (2)(i)
    )
    defined_for = StateCode.KY

    adds = ["itemized_deductions_less_salt"]
    subtracts = ["medical_expense_deduction"]



from policyengine_us.model_api import *
from policyengine_us.variables.household.demographic.person.immigration_status import (
    ImmigrationStatus,
)


class immigration_status_str(Variable):
    value_type = Enum
    possible_values = ImmigrationStatus
    default_value = ImmigrationStatus.CITIZEN
    entity = Person
    label = "ImmigrationStatus enumeration type as an all-upper-case string"
    definition_period = YEAR



from policyengine_us.model_api import *


class or_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "OR deductions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.oregonlegislature.gov/bills_laws/ors/ors316.html",  # 316.695 (1)(c)
        "https://www.oregon.gov/dor/forms/FormsPubs/form-or-40_101-040_2021.pdf#page=3",
    )
    defined_for = StateCode.OR

    def formula(tax_unit, period, parameters):
        return max_(
            tax_unit("or_itemized_deductions", period),
            tax_unit("or_standard_deduction", period),
        )



from policyengine_us.model_api import *


class sc_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "South Carolina taxable income"
    defined_for = StateCode.SC
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.sc.gov/forms-site/Forms/IITPacket_2022.pdf#page=33"
    )

    def formula(tax_unit, period, parameters):
        taxable_income = tax_unit("taxable_income", period)
        additions = tax_unit("sc_additions", period)
        subtractions = tax_unit("sc_subtractions", period)
        return max_(0, taxable_income + additions - subtractions)



from policyengine_us.model_api import *


class md_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD AGI"
    unit = USD
    definition_period = YEAR
    reference = "https://govt.westlaw.com/mdc/Browse/Home/Maryland/MarylandCodeCourtRules?guid=NAE804370A64411DBB5DDAC3692B918BC&transitionType=Default&contextData=%28sc.Default%29"
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        us_agi = tax_unit("adjusted_gross_income", period)
        md_adds = tax_unit("md_total_additions", period)
        md_subs = tax_unit("md_total_subtractions", period)
        return max_(0, us_agi + md_adds - md_subs)



from policyengine_us.model_api import *


class income_decile(Variable):
    label = "income decile"
    documentation = "Decile of household net income. Households are sorted by disposable income, and then divided into 10 equally-populated groups."
    entity = Person
    definition_period = YEAR
    value_type = int

    def formula(person, period, parameters):
        return person.household("household_income_decile", period)



from policyengine_us.model_api import *


class nc_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "North Carolina taxable income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NC

    adds = ["adjusted_gross_income", "nc_additions"]
    subtracts = ["nc_deductions"]



from policyengine_us.model_api import *


class is_in_public_housing(Variable):
    value_type = bool
    entity = Household
    label = "Whether the household is in public housing"
    definition_period = YEAR



from policyengine_us.model_api import *


class nyc_school_credit_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "NYC income used for school tax credit"
    unit = USD
    definition_period = YEAR
    defined_for = "in_nyc"
    adds = ["adjusted_gross_income"]

    # NYC School Tax Credit has its own definition of income.
    # "Income, for purposes of determining your New York City school tax credit, means your recomputed federal adjusted gross income from
    # Form IT-201, line 19a, minus distributions from an individual retirement account and an individual retirement annuity, from Form IT-201,
    # line 9, if they were included in your recomputed federal adjusted gross income."
    # https://www.tax.ny.gov/pdf/2022/printable-pdfs/inc/it201i-2022.pdf#page=26.

    # Recomputed federal AGI is only different from federal AGI if you were required to report any adjustments due to Decoupling from the IRC.
    # If we choose to ignore this possibility for now, then I think School Tax Credit Income = Federal AGI
    # https://www.tax.ny.gov/pdf/current_forms/it/it558i.pdf



from policyengine_us.model_api import *


class il_aabd_countable_unearned_income(Variable):
    value_type = float
    entity = Person
    definition_period = MONTH
    label = "Illinois Aid to the Aged, Blind or Disabled (AABD) countable unearned income"
    reference = (
        "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-113.120",
    )
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        p = parameters(period).gov.states.il.dhs.aabd.income.exemption
        gross_unearned_income = person("il_aabd_gross_unearned_income", period)
        return max_(gross_unearned_income - p.flat, 0)



from policyengine_us.model_api import *


class ma_part_b_taxable_income_before_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA Part B taxable income before exemption"
    unit = USD
    definition_period = YEAR
    reference = "https://www.mass.gov/doc/2021-form-1-massachusetts-resident-income-tax-return/download"
    defined_for = StateCode.MA

    def formula(tax_unit, period, parameters):
        part_b_agi = tax_unit("ma_part_b_agi", period)
        part_b_deductions = tax_unit(
            "ma_part_b_taxable_income_deductions", period
        )
        return max_(0, part_b_agi - part_b_deductions)



from policyengine_us.model_api import *


class unreimbursed_business_employee_expenses(Variable):
    value_type = float
    entity = Person
    label = "Unreimbursed business employee expenses"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class ar_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arkansas Child and Dependent Care Credit"
    unit = USD
    documentation = "https://codes.findlaw.com/ar/title-26-taxation/ar-code-sect-26-51-502/"
    definition_period = YEAR
    defined_for = StateCode.AR

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ar.tax.income.credits.cdcc
        # Arkansas matches the federal credit taken
        cdcc = tax_unit("cdcc", period)
        return cdcc * p.match



from policyengine_us.model_api import *


class ia_prorate_fraction(Variable):
    value_type = float
    entity = Person
    label = "Iowa joint amount proration fraction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2022-01/IA1040%2841-001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2022IA1040%2841001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf"
    )
    defined_for = StateCode.IA

    def formula(person, period, parameters):
        net_income = person("ia_net_income", period)
        total_net_income = person.tax_unit.sum(net_income)
        # avoid divide-by-zero warnings when using where() function
        fraction = np.zeros_like(total_net_income)
        mask = total_net_income != 0
        fraction[mask] = net_income[mask] / total_net_income[mask]
        # if no net income, then assign entirely to head.
        return where(
            total_net_income == 0,
            person("is_tax_unit_head", period),
            fraction,
        )



from policyengine_us.model_api import *


class va_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = "https://www.tax.virginia.gov/sites/default/files/taxforms/individual-income-tax/2021/schedule-2021.pdf"
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        p_va = parameters(period).gov.states.va.tax.income.deductions.itemized

        itm_deds_less_salt = tax_unit("itemized_deductions_less_salt", period)
        uncapped_state_and_local_tax = tax_unit(
            "state_and_local_sales_or_income_tax", period
        )
        uncapped_property_taxes = add(tax_unit, period, ["real_estate_taxes"])
        va_itm_deds = (
            itm_deds_less_salt
            + uncapped_property_taxes
            + uncapped_state_and_local_tax
        )

        federal_agi = tax_unit("adjusted_gross_income", period)

        filing_status = tax_unit("filing_status", period)
        applicable_amount = p_va.applicable_amount[filing_status]

        # Output if deductions are not limited
        itemized_ded_reduced_by_uncapped_state_and_local = max_(
            va_itm_deds - uncapped_state_and_local_tax, 0
        )
        # Output if deductions are limited
        selected_ded_reduced_by_reduced_state_and_local = tax_unit(
            "va_reduced_itemized_deductions", period
        )
        return where(
            federal_agi > applicable_amount,
            selected_ded_reduced_by_reduced_state_and_local,
            itemized_ded_reduced_by_uncapped_state_and_local,
        )



from policyengine_us.model_api import *


class charitable_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Charitable deduction"
    unit = USD
    documentation = "Deduction from taxable income for charitable donations."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/170"

    def formula(tax_unit, period, parameters):
        cash_donations = add(tax_unit, period, ["charitable_cash_donations"])
        non_cash_donations = add(
            tax_unit, period, ["charitable_non_cash_donations"]
        )
        positive_agi = tax_unit("positive_agi", period)
        p = parameters(period).gov.irs.deductions.itemized.charity
        capped_non_cash_donations = min_(
            non_cash_donations, p.ceiling.non_cash * positive_agi
        )

        total_cap = p.ceiling.all * positive_agi
        if p.floor.applies:
            deduction_floor = p.floor.amount * positive_agi
            reduced_non_cash_donations = max_(
                non_cash_donations - deduction_floor, 0
            )
            capped_reduced_non_cash_donations = min_(
                reduced_non_cash_donations, p.ceiling.non_cash * positive_agi
            )

            remaining_floor = max_(deduction_floor - non_cash_donations, 0)
            reduced_cash_donations = max_(cash_donations - remaining_floor, 0)
            return min_(
                capped_reduced_non_cash_donations + reduced_cash_donations,
                total_cap,
            )
        return min_(capped_non_cash_donations + cash_donations, total_cap)



from policyengine_us.model_api import *
from policyengine_us.variables.household.demographic.geographic.ucgid.ucgid_enum import (
    UCGID,
)


class ucgid_str(Variable):
    value_type = str
    entity = Household
    label = "UCGID (string)"
    documentation = "UCGID variable, stored as a string"
    definition_period = YEAR

    def formula(household, period, parameters):
        import numpy as np

        ucgid_enum_names = household("ucgid", period).decode_to_str()

        # Convert each enum name to its hierarchical codes
        result = []
        for enum_name in ucgid_enum_names:
            # Get the enum instance from its name
            ucgid_enum = UCGID[enum_name]

            # Get all hierarchical codes and join with commas
            hierarchical_codes = ucgid_enum.get_hierarchical_codes()
            result.append(",".join(hierarchical_codes))

        return np.array(result)



from policyengine_us.model_api import *


class dc_tanf_standard_payment(Variable):
    value_type = float
    entity = SPMUnit
    label = (
        "DC Temporary Assistance for Needy Families (TANF) standard payment"
    )
    unit = USD
    definition_period = MONTH
    reference = (
        "https://code.dccouncil.gov/us/dc/council/code/sections/4-205.52"
    )
    defined_for = StateCode.DC

    def formula(spm_unit, period, parameters):
        unit_size = spm_unit("dc_tanf_assistance_unit_size", period)
        p = parameters(period).gov.states.dc.dhs.tanf.standard_payment
        capped_unit_size = min_(unit_size, p.max_unit_size)
        return p.amount[capped_unit_size]



from policyengine_us.model_api import *


class or_exemption_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "OR exemption credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.oregon.gov/dor/forms/FormsPubs/form-or-40-inst_101-040-1_2021.pdf#page=17"
        "https://www.oregonlegislature.gov/bills_laws/ors/ors316.html"  # Subsection 316.085
    )
    defined_for = StateCode.OR

    def formula(tax_unit, period, parameters):
        EXEMPTION_TYPES = [
            "regular",
            "severely_disabled",
            "disabled_child_dependent",
        ]
        exemptions = add(
            tax_unit,
            period,
            ["or_" + i + "_exemptions" for i in EXEMPTION_TYPES],
        )
        amount = (
            parameters(period)
            .gov.states["or"]
            .tax.income.credits.exemption.amount
        )
        return exemptions * amount



from policyengine_us.model_api import *


class snap_region_str(Variable):
    value_type = str
    entity = Household
    label = "SNAP region"
    definition_period = YEAR

    def formula(household, period):
        return household("snap_region", period).decode_to_str()



from policyengine_us.model_api import *


class household_count_people(Variable):
    value_type = int
    entity = Household
    label = "Number of people"
    definition_period = YEAR
    unit = "person"

    def formula(household, period, parameters):
        return household.nb_persons()



from policyengine_us.model_api import *


class meets_snap_abawd_work_requirements(Variable):
    value_type = bool
    entity = Person
    label = "Person is eligible for SNAP benefits via Able-Bodied Adult Without Dependents (ABAWD) work requirements"
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/cfr/text/7/273.24"

    def formula(person, period, parameters):
        p = parameters(period).gov.usda.snap.work_requirements.abawd
        age = person("monthly_age", period)
        weekly_hours_worked = person("weekly_hours_worked", period.this_year)
        # Work at least 20 hours a week
        is_working = weekly_hours_worked >= p.weekly_hours_threshold
        # Adults within an age range are exempt
        working_age_exempt = p.age_threshold.exempted.calc(age)
        # Unable to work due to a physical or mental limitation
        is_disabled = person("is_disabled", period)
        # Parent or other member of a household with responsibility for a dependent child under certain age
        is_dependent = person("is_tax_unit_dependent", period)
        is_qualifying_child = age < p.age_threshold.dependent
        is_parent = person("is_parent", period)
        has_child = person.spm_unit.any(is_dependent & is_qualifying_child)
        exempt_parent = is_parent & has_child
        # Exempted from the general work requirements
        meets_snap_general_work_requirements = person(
            "meets_snap_general_work_requirements", period
        )
        # Pregnant
        is_pregnant = person("is_pregnant", period)
        # Homeless
        is_homeless = person.household("is_homeless", period)
        # A veteran
        is_veteran = person("is_veteran", period)
        # States exempt from work requirements.
        state_code = person.household("state_code", period)
        state_code_str = state_code.decode_to_str()
        is_abawd_work_requirements_exempt_state = np.isin(
            state_code_str, p.exempt_states
        )
        base_conditions = (
            is_working
            | working_age_exempt
            | is_disabled
            | exempt_parent
            | meets_snap_general_work_requirements
            | is_pregnant
            | is_abawd_work_requirements_exempt_state
        )
        if p.in_effect:
            return base_conditions
        return base_conditions | is_homeless | is_veteran



from policyengine_us.model_api import *


class health_insurance_premiums(Variable):
    value_type = float
    entity = Person
    label = "Health insurance premiums"
    unit = USD
    definition_period = YEAR

    adds = [
        "health_insurance_premiums_without_medicare_part_b",
        "medicare_part_b_premiums",
    ]



from policyengine_us.model_api import *


class dc_ccsp_qualified_need_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for DC Child Care Subsidy Program (CCSP) due to qualified need"
    definition_period = MONTH
    defined_for = StateCode.DC
    reference = "https://osse.dc.gov/sites/default/files/dc/sites/osse/publication/attachments/DC%20Child%20Care%20Subsidy%20Program%20Policy%20Manual.pdf#page=8"

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.dc.dhs.ccsp.age_threshold
        person = spm_unit.members
        age = person("monthly_age", period)
        # child age < 19 and disabled
        is_dependent = person("is_tax_unit_dependent", period)
        is_disabled = person("is_disabled", period)
        age_eligible = age < p.disabled_child
        has_disabled_child = spm_unit.any(
            is_dependent & is_disabled & age_eligible
        )
        # Income test waived when parent is disabled | child is homeless  | parent is teen parent (age <= 19)
        income_test_waived = spm_unit("dc_ccsp_income_test_waived", period)
        # parent is age >= 62 or get social_security_disability or ssi
        is_elderly = age >= p.elderly
        received_ssdi_or_ssi = (
            add(spm_unit, period, ["social_security_disability", "ssi"]) > 0
        )
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        has_elderly_parent = (
            spm_unit.any(is_head_or_spouse & is_elderly) | received_ssdi_or_ssi
        )

        return has_disabled_child | income_test_waived | has_elderly_parent



from policyengine_us.model_api import *


class csrs_retirement_pay(Variable):
    value_type = float
    entity = Person
    label = "Civil Service Retirement System (CSRS) retirement income"
    unit = USD
    definition_period = YEAR
    documentation = (
        "Retirement income from the federal Civil Service Retirement System."
    )
    reference = "https://tax.vermont.gov/individuals/seniors-and-retirees"  # Exemption for Civil Service Retirement System (CSRS)



from policyengine_us.model_api import *


class has_qdiv_or_ltcg(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Has qualified dividends or long-term capital gains"
    documentation = "Whether this tax unit has qualified dividend income or long-term capital gains income"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        # negatives amounts cannot offset other income sources
        INCOME_SOURCES = [
            "loss_limited_net_capital_gains",
            "net_capital_gains",
            "long_term_capital_gains",
            "non_sch_d_capital_gains",
            "qualified_dividend_income",
        ]
        return np.any(
            [
                add(tax_unit, period, [income_source]) > 0
                for income_source in INCOME_SOURCES
            ]
        )



from policyengine_us.model_api import *


class high_efficiency_electric_home_rebate_percent_covered(Variable):
    value_type = float
    entity = TaxUnit
    label = "Percent of expenditures covered by high electricity home rebate"
    definition_period = YEAR
    unit = USD

    def formula(tax_unit, period, parameters):
        income_ami_ratio = tax_unit("tax_unit_income_ami_ratio", period)
        p = parameters(period).gov.doe.high_efficiency_electric_home_rebate
        return p.percent_covered.calc(income_ami_ratio, right=True)



from policyengine_us.model_api import *


class is_incapable_of_self_care(Variable):
    value_type = bool
    entity = Person
    label = "Incapable of self-care"
    documentation = "Whether this person is physically or mentally incapable of caring for themselves."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/21"



from policyengine_us.model_api import *


class air_sealing_ventilation_expenditures(Variable):
    value_type = float
    entity = TaxUnit
    label = "Expenditures on air sealing and ventilation"
    unit = USD
    definition_period = YEAR
    # NB: Combined with insulation expenditures for the rebate.
    reference = "https://www.democrats.senate.gov/imo/media/doc/inflation_reduction_act_of_2022.pdf#page=585"



# policyengine_us/variables/gov/ssa/ssi/eligibility/income/deemed/from_ineligible_spouse/ssi_income_deemed_from_ineligible_spouse.py

from policyengine_us.model_api import *
from policyengine_us.variables.gov.ssa.ssi.eligibility.income._apply_ssi_exclusions import (
    _apply_ssi_exclusions,
)


class ssi_income_deemed_from_ineligible_spouse(Variable):
    value_type = float
    entity = Person
    label = "SSI income (deemed from ineligible spouse)"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/cfr/text/20/416.1163"
    documentation = """
    Spousal deeming: 
      1) If leftover spouse income <= (coupleFBR - indivFBR), then 0 is deemed.
      2) Otherwise, spouse's deemed = (couple combined countable) - (individual alone countable).
         This yields 816 for 1986 Example 3 and 12000 for the 2025 test.
    """
    defined_for = "is_ssi_eligible_individual"
    unit = USD

    def formula(person, period, parameters):
        # 1. Sum leftover spouse earned/unearned (post-child allocations).
        spouse_earned = person(
            "ssi_earned_income_deemed_from_ineligible_spouse", period
        )
        spouse_unearned = person(
            "ssi_unearned_income_deemed_from_ineligible_spouse", period
        )
        leftover_spouse = spouse_earned + spouse_unearned

        # 2. Compare leftover to FBR difference
        p = parameters(period).gov.ssa.ssi.amount
        diff = (p.couple - p.individual) * MONTHS_IN_YEAR
        if_leftover_exceeds = leftover_spouse > diff

        # 3. If leftover <= diff => no deeming
        #    If leftover > diff => difference of incomes approach
        #    (couple combined) - (individual alone).
        #   (a) individual's own income
        indiv_earned = person("ssi_earned_income", period)
        indiv_unearned = person("ssi_unearned_income", period)
        alone_countable = _apply_ssi_exclusions(
            indiv_earned, indiv_unearned, parameters, period
        )

        #   (b) couple combined income
        couple_countable = _apply_ssi_exclusions(
            indiv_earned + spouse_earned,
            indiv_unearned + spouse_unearned,
            parameters,
            period,
        )

        deemed_if_exceeds = max_(0, couple_countable - alone_countable)
        return if_leftover_exceeds * deemed_if_exceeds



from policyengine_us.model_api import *


class ma_liheap_eligible_subsidized_housing(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Massachusetts LIHEAP eligible subsidized housing"
    definition_period = YEAR
    defined_for = StateCode.MA
    reference = "https://liheapch.acf.gov/tables/subsidize.htm#MA"

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.ma.doer.liheap.eligibility
        income = spm_unit("ma_liheap_income", period)
        rent_threshold = income * p.rent_threshold_subsidized_housing
        total_rent = add(spm_unit, period, ["rent"])
        # Eligible if monthly rent is more than a certain fraction of income
        rent_eligible = total_rent > rent_threshold
        is_subsidized = spm_unit("receives_housing_assistance", period)
        heat_in_rent = spm_unit("heat_expense_included_in_rent", period)

        return (is_subsidized & heat_in_rent & rent_eligible) | (
            is_subsidized & ~heat_in_rent
        )



from policyengine_us.model_api import *


class dc_liheap_payment(Variable):
    value_type = float
    entity = SPMUnit
    label = "DC LIHEAP payment"
    unit = USD
    definition_period = YEAR
    defined_for = "dc_liheap_eligible"
    reference = "https://doee.dc.gov/sites/default/files/dc/sites/doee/service_content/attachments/DOEE%20FY24%20LIHEAP_REGULAR_Benefits_Table-Matrix.pdf"

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.dc.doee.liheap.payment
        housing_type = spm_unit("dc_liheap_housing_type", period)
        heating_type = spm_unit("dc_liheap_heating_type", period)
        income_level = spm_unit("dc_liheap_income_level", period)
        size = spm_unit("spm_unit_size", period)
        capped_size = clip(size, 1, 4)

        electricity = heating_type == heating_type.possible_values.ELECTRICITY
        uncapped_electricity_payment = p.electricity[housing_type][
            income_level
        ][capped_size]
        electricity_expense = spm_unit(
            "pre_subsidy_electricity_expense", period
        )
        electricity_payment = min_(
            uncapped_electricity_payment, electricity_expense
        )

        gas = heating_type == heating_type.possible_values.GAS
        uncapped_gas_payment = p.gas[housing_type][income_level][capped_size]
        gas_expense = spm_unit("gas_expense", period)
        gas_payment = min_(uncapped_gas_payment, gas_expense)

        heat_in_rent = (
            heating_type == heating_type.possible_values.HEAT_IN_RENT
        )

        oil = heating_type == heating_type.possible_values.OIL
        oil_expense = spm_unit("fuel_oil_expense", period)
        oil_payment = min_(p.oil, oil_expense)

        return select(
            [electricity, gas, heat_in_rent, oil],
            [
                electricity_payment,
                gas_payment,
                p.heat_in_rent,
                oil_payment,
            ],
            default=0,
        )



from policyengine_us.model_api import *


class utilities_included_in_rent(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Whether utilities are included in rent payments"
    definition_period = YEAR



from policyengine_us.model_api import *


class ut_retirement_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Utah retirement credit"
    unit = USD
    definition_period = YEAR
    defined_for = "ut_claims_retirement_credit"

    adds = ["ut_retirement_credit_max"]



from policyengine_us.model_api import *


class de_standard_deduction_indv(Variable):
    value_type = float
    entity = Person
    label = "Delaware standard deduction when married couples are filing separately"
    unit = USD
    definition_period = YEAR
    reference = "https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf#page=8"
    defined_for = StateCode.DE

    adds = [
        "de_base_standard_deduction_indv",
        "de_additional_standard_deduction",
    ]



from policyengine_us.model_api import *


class auto_loan_balance(Variable):
    value_type = float
    entity = Household
    label = "auto loan total balance"
    unit = USD
    definition_period = YEAR
    uprating = "gov.bls.cpi.cpi_u"



from policyengine_us.model_api import *


class mi_interest_dividends_capital_gains_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Michigan interest, dividends, and capital gains deduction"
    unit = USD
    definition_period = YEAR
    documentation = "Michigan interest, dividends, and capital gains deduction of qualifying age."
    reference = (
        "http://legislature.mi.gov/doc.aspx?mcl-206-30",  # (1)(p)
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf#page=16",
    )
    defined_for = "mi_interest_dividends_capital_gains_deduction_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.mi.tax.income.deductions.interest_dividends_capital_gains

        # Core deduction based on filing status.
        filing_status = tax_unit("filing_status", period)
        person = tax_unit.members

        # Senior citizens may subtract interest, dividends, and capital gains included in AGI.
        income = add(tax_unit, period, p.income_types)
        # The maximum amount of the deduction must be reduced by any deduction for
        # Military (including Michigan National Guard) retirement benefits
        # Public and private retirement and pension benefits
        # Amount used for the federal credit for the elderly and totally and permanently disabled
        reductions_pay = add(
            person,
            period,
            ["military_retirement_pay", "taxable_pension_income"],
        )
        elderly_disabled_credit = tax_unit("elderly_disabled_credit", period)

        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)

        reductions = (
            tax_unit.sum(reductions_pay * is_head_or_spouse)
            + elderly_disabled_credit
        )
        reduced_amount = max_(0, p.amount[filing_status] - reductions)

        return min_(reduced_amount, income)



from policyengine_us.model_api import *


class childcare_hours_per_week(Variable):
    value_type = float
    entity = Person
    label = "Child care hours per week"
    definition_period = YEAR
    unit = "hour"

    def formula(person, period, parameters):
        days_per_week = person("childcare_days_per_week", period)
        hours_per_day = person("childcare_hours_per_day", period)
        return days_per_week * hours_per_day



from policyengine_us.model_api import *


class sc_military_deduction_survivors(Variable):
    value_type = float
    entity = Person
    label = "South Carolina military retirement deduction for survivors"
    defined_for = StateCode.SC
    unit = USD
    reference = (
        "https://www.scstatehouse.gov/code/t12c006.php",  # SECTION 12-6-1171(C)
        "https://dor.sc.gov/forms-site/Forms/IITPacket_2021.pdf#page=17",
    )
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("military_retirement_pay_survivors", period)



from policyengine_us.model_api import *


class takes_up_medicaid_if_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Whether a random eligible person unit does not enroll in Medicaid"
    definition_period = YEAR

    def formula(person, period, parameters):
        seed = person("medicaid_take_up_seed", period)
        takeup_rate = parameters(period).gov.hhs.medicaid.takeup_rate
        return seed < takeup_rate



from policyengine_us.model_api import *


class pell_grant_dependent_allowances(Variable):
    value_type = float
    entity = Person
    unit = USD
    label = "Pell Grant dependent allowances"
    definition_period = YEAR

    def formula(person, period, parameters):
        other_allowances = person(
            "pell_grant_dependent_other_allowances", period
        )
        ipa = parameters(period).gov.ed.pell_grant.dependent.ipa
        head_available_income = person(
            "pell_grant_head_available_income", period
        )
        allowances_from_head = -min_(head_available_income, 0)
        return ipa + allowances_from_head + other_allowances



from policyengine_us.model_api import *


class ma_tafdc_pregnancy_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Eligible for Massachusetts Temporary Assistance for Families with Dependent Children (TAFDC) due to pregnancy"
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-703-210"
    )
    defined_for = StateCode.MA

    def formula(person, period, parameters):
        is_pregnant = person("is_pregnant", period)
        age = person("age", period)
        current_pregnancy_month = person("current_pregnancy_month", period)
        p = parameters(period).gov.states.ma.dta.tcap.tafdc.eligibility
        age_limit = person("ma_tafdc_age_limit", period)
        age_eligible = age < age_limit
        # Eligible if pregnant and are in the 5th month or later (due in 120 days)
        months_eligible = is_pregnant & (
            current_pregnancy_month >= p.pregnancy_month
        )
        # Eligible if age under 20 and meet teen parent school attendance requirements
        is_in_secondary_school = person("is_in_secondary_school", period)
        teen_pregnancy_eligible = (
            is_pregnant & age_eligible & is_in_secondary_school
        )
        return teen_pregnancy_eligible | months_eligible



from policyengine_us.model_api import *


class ut_income_tax_exempt(Variable):
    value_type = bool
    entity = TaxUnit
    label = "exempt from Utah income tax"
    unit = USD
    documentation = "Form TC-40, line 21"
    definition_period = YEAR
    defined_for = StateCode.UT
    reference = "https://le.utah.gov/xcode/Title59/Chapter10/59-10-S104.1.html?v=C59-10-S104.1_1800010118000101"  # (2)

    def formula(tax_unit, period, parameters):
        federal_agi = tax_unit("adjusted_gross_income", period)
        basic_standard_deduction = tax_unit("basic_standard_deduction", period)
        return federal_agi <= basic_standard_deduction



from policyengine_us.model_api import *


class me_personal_exemption_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maine personal exemption deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.ME

    def formula(tax_unit, period, parameters):
        # First get their Maine AGI.
        me_agi = tax_unit("me_agi", period)

        # Get filing status.
        filing_status = tax_unit("filing_status", period)

        # Then get the Maine Personal Credit part of the parameter tree.
        p = parameters(
            period
        ).gov.states.me.tax.income.deductions.personal_exemption

        # Calculate the Personal Exemption phase-out parameters based on
        # filing status.
        phaseout_start = p.phaseout.start[filing_status]
        excess = max_(me_agi - phaseout_start, 0)  # Line 3
        phaseout_width = p.phaseout.width[filing_status]  # Line 4
        phaseout_percent = min_(1, excess / phaseout_width)  # Line 5

        # Get their Maine personal exemptions (line 6).
        exemptions = tax_unit("head_spouse_count", period)
        max_amount = exemptions * p.amount

        # Calculate the phaseout amount (line 7).
        phaseout_amount = phaseout_percent * max_amount

        return max_amount - phaseout_amount  # Line 8



from policyengine_us.model_api import *


class dividend_income_reduced_by_investment_income(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Dividend income reduced by investment income"
    documentation = "IRS Form 1040 Schedule D worksheet (part 1 of 6)"
    unit = USD

    def formula(tax_unit, period, parameters):
        qualified_dividend_income = add(
            tax_unit, period, ["qualified_dividend_income"]
        )
        investment_income = tax_unit("investment_income_form_4952", period)
        # dwks04 always assumed to be zero
        floored_investment_income = max_(0, investment_income)
        return max_(0, qualified_dividend_income - floored_investment_income)



from policyengine_us.model_api import *


class mi_retirement_benefits_deduction_tier_three_ss_exempt_retired_eligible_people(
    Variable
):
    value_type = int
    entity = TaxUnit
    label = "Eligible for the Michigan tier three retired retirement benefits deduction"
    definition_period = YEAR
    reference = (
        "http://legislature.mi.gov/doc.aspx?mcl-206-30",  # (9)(c)
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf#page=18",
        "https://www.michigan.gov/taxes/iit/retirement-and-pension-benefits",
    )
    defined_for = StateCode.MI

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.mi.tax.income.deductions.retirement_benefits.tier_three.ss_exempt.retired

        person = tax_unit.members
        # Recipients should received retirement benefits from SSA exempt employment
        # and were retired before qualifying year
        retirement_eligible = (
            person("year_of_retirement", period) <= p.retirement_year
        ) & (person("year_of_retirement", period) > 0)

        has_retirement_benefits_from_ss_exempt_employment = (
            person("retirement_benefits_from_ss_exempt_employment", period) > 0
        )

        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)

        eligible_people = (
            has_retirement_benefits_from_ss_exempt_employment
            * retirement_eligible
            * is_head_or_spouse
        )

        return tax_unit.sum(eligible_people)



from policyengine_us.model_api import *


class eitc_child_count(Variable):
    value_type = int
    entity = TaxUnit
    label = "EITC-qualifying children"
    unit = USD
    documentation = "Number of children qualifying as children for the EITC."
    definition_period = YEAR
    reference = (
        # IRC 32(c)(3)(D)(i) specifies qualifying children need a taxpayer identification number.
        "https://www.law.cornell.edu/uscode/text/26/32#c_3_D_i",
    )

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        is_child = person("is_child_dependent", period)
        meets_eitc_identification_requirements = person(
            "meets_eitc_identification_requirements", period
        )
        eligible_child = is_child & meets_eitc_identification_requirements
        # Only child who meets EITC identification requirements can claim benefit
        return tax_unit.sum(eligible_child)



from policyengine_us.model_api import *


class de_base_standard_deduction_indv(Variable):
    value_type = float
    entity = Person
    label = "Delaware base standard deduction when married couples are filing separately"
    unit = USD
    definition_period = YEAR
    reference = "https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-02_Instructions.pdf#page=8"
    defined_for = StateCode.DE

    def formula(person, period, parameters):
        filing_status = person.tax_unit(
            "state_filing_status_if_married_filing_separately_on_same_return",
            period,
        )
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        p = parameters(period).gov.states.de.tax.income.deductions.standard
        return head_or_spouse * p.amount[filing_status]



from policyengine_us.model_api import *


class ma_part_b_gross_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA Part B gross income"
    unit = USD
    definition_period = YEAR
    reference = "https://www.mass.gov/info-details/mass-general-laws-c62-ss-2"
    defined_for = StateCode.MA

    def formula(tax_unit, period, parameters):
        ma_gross_income = tax_unit("ma_gross_income", period)
        ma_part_a_gross_income = tax_unit("ma_part_a_gross_income", period)
        ma_part_c_gross_income = tax_unit("ma_part_c_gross_income", period)
        return max_(
            0,
            ma_gross_income - ma_part_a_gross_income - ma_part_c_gross_income,
        )



from policyengine_us.model_api import *


class mi_household_resources(Variable):
    value_type = float
    entity = TaxUnit
    label = "Michigan household resources"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MI
    reference = "https://law.justia.com/codes/michigan/2022/chapter-206/statute-act-281-of-1967/division-281-1967-1/division-281-1967-1-9/section-206-508/"
    adds = "gov.states.mi.tax.income.household_resources"
    subtracts = ["health_insurance_premiums", "above_the_line_deductions"]



from policyengine_us.model_api import *


class exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Exemptions"
    definition_period = YEAR
    documentation = "Personal exemptions amount after phase-out"
    unit = USD

    def formula(tax_unit, period, parameters):
        # calculate exemptions amount before phase-out
        exemptions = tax_unit("exemptions_count", period)
        p = parameters(period).gov.irs.income.exemption
        amount = exemptions * p.amount
        # calculate exemptions amount after phase-out
        filing_status = tax_unit("filing_status", period)
        phase_out_start_agi = p.phase_out.start[filing_status]
        agi = tax_unit("adjusted_gross_income", period)
        excess_agi = max_(0, agi - phase_out_start_agi)
        phase_out_step_size = p.phase_out.step_size[filing_status]
        steps = excess_agi / phase_out_step_size
        phase_out_fraction = steps * p.phase_out.rate
        return max_(0, amount * (1 - phase_out_fraction))



from policyengine_us.model_api import *


class ca_state_supplement_eligible_person(Variable):
    value_type = bool
    entity = Person
    label = "California SSI state supplement eligible person"
    definition_period = MONTH
    defined_for = StateCode.CA
    reference = "https://law.justia.com/codes/california/code-wic/division-9/part-3/chapter-3/article-4/"

    def formula(person, period, parameters):
        meets_resource_test = person("meets_ssi_resource_test", period)
        aged_blind_disabled = person("is_ssi_aged_blind_disabled", period)
        is_qualified_noncitizen = person("is_ssi_qualified_noncitizen", period)
        immigration_status = person("immigration_status", period)
        is_citizen = (
            immigration_status == immigration_status.possible_values.CITIZEN
        )
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        return (
            meets_resource_test
            & aged_blind_disabled
            & (is_qualified_noncitizen | is_citizen)
            & is_head_or_spouse
        )



from policyengine_us.model_api import *


def create_abolish_snap_deductions() -> Reform:
    class snap_deductions(Variable):
        value_type = float
        entity = SPMUnit
        label = "SNAP income deductions"
        unit = USD
        documentation = "Deductions made from gross income for SNAP benefits"
        definition_period = MONTH
        reference = "https://www.law.cornell.edu/uscode/text/7/2014#e"

        def formula(spm_unit, period, parameters):
            return 0

    class reform(Reform):
        def apply(self):
            self.update_variable(snap_deductions)

    return reform


def create_abolish_snap_deductions_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_abolish_snap_deductions()

    p = parameters(period).gov.contrib.snap.abolish_deductions

    if p.in_effect:
        return create_abolish_snap_deductions()
    else:
        return None


abolish_snap_deductions = create_abolish_snap_deductions_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class il_cta_military_service_pass_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Eligible for the Illinois Chicago Transit Authority military service pass"
    definition_period = YEAR
    defined_for = StateCode.IL
    reference = "https://www.transitchicago.com/military/"

    def formula(person, period, parameters):
        is_in_active_duty = person("military_basic_pay", period) > 0
        disabled_veterans = person(
            "is_fully_disabled_service_connected_veteran", period
        )
        return is_in_active_duty | disabled_veterans



from policyengine_us.model_api import *


class id_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Idaho deductions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://legislature.idaho.gov/statutesrules/idstat/Title63/T63CH30/SECT63-3022/",  # (j)
        "https://tax.idaho.gov/wp-content/uploads/forms/EIN00046/EIN00046_03-01-2023.pdf#page=8",
    )
    defined_for = StateCode.ID

    def formula(tax_unit, period, parameters):
        itm_ded = tax_unit("id_itemized_deductions", period)
        std_ded = tax_unit("standard_deduction", period)
        # We do not model qualified business income deduction for Idaho
        return max_(itm_ded, std_ded)



from policyengine_us.model_api import *


class taxable_income_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Taxable income deductions"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        itemizes = tax_unit("tax_unit_itemizes", period)
        deductions_if_itemizing = tax_unit(
            "taxable_income_deductions_if_itemizing", period
        )
        deductions_if_not_itemizing = tax_unit(
            "taxable_income_deductions_if_not_itemizing", period
        )
        return where(
            itemizes, deductions_if_itemizing, deductions_if_not_itemizing
        )



from policyengine_us.model_api import *


class heating_cooling_expense(Variable):
    value_type = float
    entity = SPMUnit
    label = "Heating and cooling expense"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class dc_eitc_without_qualifying_child(Variable):
    value_type = float
    entity = TaxUnit
    label = "DC EITC without qualifying children"
    unit = USD
    definition_period = YEAR
    reference = "https://code.dccouncil.gov/us/dc/council/code/sections/47-1806.04"  # (f)
    defined_for = StateCode.DC

    def formula(tax_unit, period, parameters):
        # calculate US EITC amount before phase-out
        us_eligible = tax_unit("eitc_eligible", period)
        us_eitc = us_eligible * tax_unit("eitc_phased_in", period)
        # phase out us_eitc for income above DC phase-out start threshold
        earnings = tax_unit("tax_unit_earned_income", period)
        us_agi = tax_unit("adjusted_gross_income", period)
        greater_of = max_(earnings, us_agi)
        dc = parameters(period).gov.states.dc.tax.income.credits
        excess = max_(0, greater_of - dc.eitc.without_children.phase_out.start)
        dc_phase_out_amount = excess * dc.eitc.without_children.phase_out.rate
        return max_(0, us_eitc - dc_phase_out_amount)



from policyengine_us.model_api import *


class takes_up_snap_if_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Whether a random eligible SPM unit does not claim SNAP"
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        seed = spm_unit("snap_take_up_seed", period)
        takeup_rate = parameters(period).gov.usda.snap.takeup_rate
        return seed < takeup_rate



from policyengine_us.model_api import *


class ms_total_exemptions_joint(Variable):
    value_type = float
    entity = Person
    label = "Mississippi total exemptions when married couples file jointly"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MS
    reference = "https://www.dor.ms.gov/sites/default/files/Forms/Individual/80100221.pdf#page=6"

    def formula(person, period, parameters):
        total_exemptions = person.tax_unit("ms_total_exemptions", period)
        is_head = person("is_tax_unit_head", period)
        return is_head * total_exemptions



from policyengine_us.model_api import *


class is_ssi_ineligible_spouse(Variable):
    value_type = bool
    entity = Person
    label = "Is an SSI-ineligible spouse"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/42/1382c#b"

    def formula(person, period, parameters):
        # Check if person is a spouse (either tax unit spouse or head)
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)

        # Check SSI eligibility status
        eligible = person("is_ssi_aged_blind_disabled", period)
        eligible_spouse = person("is_ssi_eligible_spouse", period)

        # A person is an ineligible spouse if they:
        # 1. Are a spouse (either tax unit spouse or head)
        # 2. Are not SSI eligible themselves
        # 3. Are not an SSI eligible spouse
        return is_head_or_spouse & ~eligible_spouse & ~eligible



"""IRS uprating parameter setup.

This module imports the unified uprating extension functionality.
The actual extension logic is in parameters/uprating_extensions.py.
"""

from policyengine_us.parameters.uprating_extensions import (
    set_all_uprating_parameters as set_irs_uprating_parameter,
)

# For backward compatibility, we keep the same function name
# but it now calls the unified function that handles all uprating parameters



from policyengine_us.model_api import *


class ia_income_tax_before_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Iowa income tax before refundable credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2022-01/IA1040%2841-001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf#page=60"
        "https://tax.iowa.gov/sites/default/files/2023-01/2022IA1040%2841001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf#page=60"
    )
    defined_for = StateCode.IA

    def formula(tax_unit, period, parameters):
        tax_before = tax_unit("ia_income_tax_before_credits", period)
        nonref_credits = tax_unit("ia_nonrefundable_credits", period)
        is_tax_exempt = tax_unit("ia_is_tax_exempt", period)
        tax_after = ~is_tax_exempt * max_(0, tax_before - nonref_credits)
        tax_reduced = tax_unit("ia_reduced_tax", period)
        return min_(tax_after, tax_reduced)



from policyengine_us.model_api import *


class ky_family_size_tax_credit_rate(Variable):
    value_type = float
    entity = TaxUnit
    label = "Kentucky family size tax credit rate"
    unit = "/1"
    definition_period = YEAR
    reference = (
        "https://apps.legislature.ky.gov/law/statutes/statute.aspx?id=49188"
    )
    defined_for = StateCode.KY

    def formula(tax_unit, period, parameters):
        income = tax_unit("ky_modified_agi", period)
        fpg = parameters(period).gov.hhs.fpg
        # This will be CONTIGUOUS_US for Kentucky.
        state_group = tax_unit.household("state_group", period)
        p1 = fpg.first_person[state_group]
        padd = fpg.additional_person[state_group]
        family_size = tax_unit("tax_unit_size", period)
        # No more than 4 people are accounted for in the credit
        p = parameters(period).gov.states.ky.tax.income.credits.family_size
        capped_family_size = min_(family_size, p.family_size_cap)
        poverty_index = p1 + padd * (capped_family_size - 1)
        share = income / poverty_index
        return p.rate.calc(share, right=True)



from policyengine_us.model_api import *


class is_hud_elderly_disabled_family(Variable):
    value_type = bool
    entity = SPMUnit
    label = "HUD elderly or disabled family"
    unit = USD
    documentation = (
        "Whether an SPM unit is deemed elderly or disabled for HUD purposes"
    )
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/cfr/text/24/5.611"

    def formula(spm_unit, period, parameters):
        hud = parameters(period).gov.hud
        person = spm_unit.members
        elderly = person("age", period) >= hud.elderly_age_threshold
        disabled = person("is_disabled", period)
        child = person("is_child", period)
        elderly_disabled_adult = (elderly | disabled) & ~child
        # Simplify to having any elderly or disabled adults.
        # Actual rule only applies to head of household or spouse.
        return spm_unit.any(elderly_disabled_adult)



from policyengine_us.model_api import *


class co_pension_subtraction_income(Variable):
    value_type = float
    entity = Person
    label = "Income for the Colorado pension and annuity subtraction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CO

    adds = "gov.states.co.tax.income.subtractions.pension.income_sources"



from policyengine_us.model_api import *


class ca_cdcc_relevant_expenses(Variable):
    value_type = float
    entity = TaxUnit
    label = "CDCC-relevant care expenses replicated to include California limitations"
    unit = USD
    definition_period = YEAR
    reference = "https://www.ftb.ca.gov/about-ftb/data-reports-plans/Summary-of-Federal-Income-Tax-Changes/index.html#PL-117-2-9631"
    defined_for = StateCode.CA

    def formula(tax_unit, period, parameters):
        year = period.start.year
        if year == 2021:
            period_adjusted = f"{year-1}-01-01"
        else:
            period_adjusted = f"{year}-01-01"

        # First, cap based on the number of eligible care receivers
        expenses = tax_unit("tax_unit_childcare_expenses", period)
        cdcc = parameters(period_adjusted).gov.irs.credits.cdcc
        count_eligible = min_(
            cdcc.eligibility.max, tax_unit("count_cdcc_eligible", period)
        )
        eligible_capped_expenses = min_(expenses, cdcc.max * count_eligible)
        # Then, cap further to the lowest earnings between the taxpayer and spouse
        return min_(
            eligible_capped_expenses,
            tax_unit("min_head_spouse_earned", period),
        )



from policyengine_us.model_api import *


class non_refundable_american_opportunity_credit_credit_limit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Non-refundable American Opportunity Credit credit limit"
    unit = USD
    documentation = "Value of the non-refundable portion of the American Opportunity Credit"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/25A#i"

    def formula(tax_unit, period, parameters):
        income_tax_before_credits = tax_unit(
            "income_tax_before_credits", period
        )
        p = parameters(
            period
        ).gov.irs.credits.education.american_opportunity_credit
        preceding_credits = add(tax_unit, period, p.preceding_credits)
        return max_(income_tax_before_credits - preceding_credits, 0)



from policyengine_us.model_api import *


class total_disability_payments(Variable):
    value_type = float
    entity = Person
    label = "Disability (total) payments"
    unit = USD
    documentation = "Wages (or payments in lieu thereof) paid to an individual for permanent and total disability"
    definition_period = YEAR



from policyengine_us.model_api import *


class is_eligible_for_fsla_overtime(Variable):
    value_type = bool
    entity = Person
    label = "is eligible for overtime pay"
    reference = "https://www.law.cornell.edu/cfr/text/29/541.600 ; https://www.law.cornell.edu/uscode/text/29/213"
    definition_period = YEAR

    def formula_2014(person, period, parameters):
        # Get the applicable threshold
        applicable_threshold = person("fsla_overtime_salary_threshold", period)

        # Check payment method and income
        is_paid_hourly = person("is_paid_hourly", period)
        employment_income = person("employment_income", period)

        # The worker is exempt if:
        # 1. Their income exceeds the threshold corresponding to their exemption category
        # 2. They're not paid hourly
        is_exempt = (employment_income >= applicable_threshold) & (
            ~is_paid_hourly
        )

        # Return True if eligible for overtime protection, False if exempt
        return ~is_exempt



from policyengine_us.model_api import *


class ne_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "NE taxable income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://revenue.nebraska.gov/files/doc/tax-forms/2021/f_1040n_booklet.pdf"
        "https://revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_5.pdf"
    )
    defined_for = StateCode.NE

    def formula(tax_unit, period, parameters):
        agi = tax_unit("ne_agi", period)
        # 2021 (2022) Form 1040N instructions on page 7 (page 8) say this:
        #   If you use the standard deduction on the federal return,
        #   you must use the Nebraska standard deduction on the
        #   Nebraska return.  All taxpayers that claimed itemized
        #   deductions on their federal return are allowed the larger
        #   of the Nebraska standard deduction or federal itemized
        #   deductions, minus state and local income taxes claimed on
        #   Federal Schedule A.
        federal_itemizer = tax_unit("tax_unit_itemizes", period)
        std_ded = tax_unit("ne_standard_deduction", period)
        itm_ded = tax_unit("ne_itemized_deductions", period)
        deduction = where(federal_itemizer, max_(itm_ded, std_ded), std_ded)
        return max_(0, agi - deduction)



from policyengine_us.model_api import *


class ca_tanf_income_limit(Variable):
    value_type = float
    entity = SPMUnit
    label = "California CalWORKs Minimum Basic Standard of Adequate Care"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CA
    reference = "http://epolicy.dpss.lacounty.gov/epolicy/epolicy/server/general/projects_responsive/ePolicyMaster/index.htm?&area=general&type=responsivehelp&ctxid=&project=ePolicyMaster#t=mergedProjects%2FCalWORKs%2FCalWORKs%2F44-212_Minimum_Basic_Standard_of_Adequate_Care%2F44-212_Minimum_Basic_Standard_of_Adequate_Care.htm%23Documentsbc-6&rhtocid=_3_1_7_20_5"

    def formula(spm_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.ca.cdss.tanf.cash.income.monthly_limit
        unit_size = spm_unit("spm_unit_size", period)
        au_size = min_(unit_size, p.max_au_size)
        additional_people = unit_size - au_size
        region1 = spm_unit.household("ca_tanf_region1", period)

        main_limit = where(
            region1, p.region1.main[au_size], p.region2.main[au_size]
        )
        increase_per_additional_person = where(
            region1, p.region1.additional, p.region2.additional
        )

        monthly_limit = (
            main_limit + increase_per_additional_person * additional_people
        )
        return monthly_limit * MONTHS_IN_YEAR



from policyengine_us.model_api import *


class is_ssi_ineligible_child(Variable):
    value_type = bool
    entity = Person
    label = "Is an SSI-ineligible child"
    definition_period = YEAR

    def formula(person, period, parameters):
        abd = person("is_ssi_aged_blind_disabled", period)
        child = person("is_child", period)
        return ~abd & child



from policyengine_us.model_api import *


class medicaid(Variable):
    value_type = float
    entity = Person
    label = "Medicaid"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/42/1396a"
    adds = ["medicaid_cost"]



from policyengine_us.model_api import *


class is_veteran(Variable):
    value_type = bool
    entity = Person
    label = "Is veteran"
    documentation = "A person who served in the active military, naval, air, or space service, and who was discharged or released therefrom under conditions other than dishonorable."
    reference = "https://www.law.cornell.edu/uscode/text/38/101"  # (2)
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("veterans_benefits", period) > 0



from policyengine_us.model_api import *


class qualified_business_income_deduction_person(Variable):
    value_type = float
    entity = Person
    label = "Qualified business income deduction for each person"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/uscode/text/26/199A#b_1"
        "https://www.irs.gov/pub/irs-prior/p535--2018.pdf"
    )

    def formula(person, period, parameters):
        # Allocate the business income deduction to each person in the tax unit
        # based on their share of per cap qualified business income deduction amount
        qbid_amt = person("qbid_amount", period)
        total_qbid_amount = person.tax_unit.sum(qbid_amt)
        total_deduction_amount = person.tax_unit(
            "qualified_business_income_deduction", period
        )
        deduction_fraction = np.zeros_like(total_qbid_amount)
        mask = total_qbid_amount != 0
        deduction_fraction[mask] = qbid_amt[mask] / total_qbid_amount[mask]
        return deduction_fraction * total_deduction_amount



from .abolish_federal_income_tax import (
    create_abolish_federal_income_tax_reform,
)
from .abolish_payroll_tax import create_abolish_payroll_tax_reform
from .reported_state_income_tax import create_reported_state_income_tax_reform



from policyengine_us.model_api import *


class ssi_marital_both_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Both members of the marital unit are eligible for SSI"
    definition_period = YEAR

    def formula(person, period, parameters):
        eligible = person("is_ssi_aged_blind_disabled", period)

        return person.marital_unit.sum(eligible) == 2



from policyengine_us.model_api import *


class wv_gross_household_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "West Virginia gross household income"
    reference = (
        "https://code.wvlegislature.gov/11-21-23/"
        "https://tax.wv.gov/Documents/TaxForms/2021/it140.pdf#page=13"
        "https://tax.wv.gov/Documents/PIT/2022/PersonalIncomeTaxFormsAndInstructions.2022.pdf#page=14"
    )
    definition_period = YEAR
    defined_for = StateCode.WV

    adds = (
        "gov.states.wv.tax.income.credits.heptc.gross_household_income.sources"
    )



from policyengine_us.model_api import *


class oh_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Ohio taxable income"
    unit = USD
    definition_period = YEAR
    reference = "https://tax.ohio.gov/static/forms/ohio_individual/individual/2021/pit-it1040-booklet.pdf"
    defined_for = StateCode.OH

    def formula(tax_unit, period, parameters):
        agi = tax_unit("oh_agi", period)
        exemptions = tax_unit("oh_personal_exemptions", period)

        return max_(0, agi - exemptions)



from policyengine_us.model_api import *


class in_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Indiana exemptions"
    unit = USD
    definition_period = YEAR
    reference = "http://iga.in.gov/legislative/laws/2021/ic/titles/006#6-3-1-3.5"  # (a)(1)-(5)
    defined_for = StateCode.IN

    adds = [
        "in_base_exemptions",
        "in_additional_exemptions",
        "in_aged_blind_exemptions",
        "in_aged_low_agi_exemptions",
        "in_adoption_exemption",
    ]



from policyengine_us.model_api import *


class household_tax(Variable):
    value_type = float
    entity = Household
    label = "total tax after refundable credits"
    documentation = "Total tax liability after refundable credits."
    unit = USD
    definition_period = YEAR
    adds = ["household_tax_before_refundable_credits"]
    subtracts = ["household_refundable_tax_credits"]



from policyengine_us.model_api import *


class ma_eaedc_eligible_dependent(Variable):
    value_type = bool
    entity = Person
    label = "Eligible dependent for the Massachusetts EAEDC dependent care deduction"
    definition_period = MONTH
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-703-200"
    )
    defined_for = StateCode.MA

    def formula(person, period, parameters):
        p = parameters(period).gov.states.ma.dta.tcap.eaedc.age_threshold
        dependent = person("is_tax_unit_dependent", period)
        age = person("monthly_age", period)
        age_eligible = age < p.dependent
        is_related_to_head_or_spouse = person(
            "is_tafdc_related_to_head_or_spouse", period
        )
        return dependent & age_eligible & ~is_related_to_head_or_spouse



from policyengine_us.model_api import *


class co_ctc_eligible_child(Variable):
    value_type = bool
    entity = Person
    label = "Colorado child tax credit eligible child"
    definition_period = YEAR
    reference = (
        "https://leg.colorado.gov/sites/default/files/2023a_1112_signed.pdf#page=2",
    )
    defined_for = StateCode.CO

    adds = "gov.states.co.tax.income.credits.ctc.eligible_child"



from numpy.typing import ArrayLike
from typing import Callable, Tuple
from policyengine_core.populations import Population
from policyengine_core.periods import Period
from policyengine_core.parameters import ParameterNode

Formula = Callable[[Tuple[Population, Period, ParameterNode]], ArrayLike]



from policyengine_us.model_api import *


class mo_adjusted_gross_income(Variable):
    value_type = float
    entity = Person
    label = "Missouri adjusted gross income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.mo.gov/forms/MO-1040%20Fillable%20Calculating_2021.pdf",
        "https://revisor.mo.gov/main/OneSection.aspx?section=143.121",
    )
    defined_for = StateCode.MO

    def formula(person, period, parameters):
        gross_income = person("irs_gross_income", period)
        # subtract federal above-the-line deductions (ALDs) by person
        # ... subtract some ALDs explicitly by person
        PERSONAL_ALDS = [
            "self_employment_tax_ald_person",
            "self_employed_health_insurance_ald_person",
            "self_employed_pension_contribution_ald_person",
        ]
        tax_unit = person.tax_unit
        ind_total_personal_alds = add(person, period, PERSONAL_ALDS)
        unit_total_personal_alds = add(tax_unit, period, PERSONAL_ALDS)
        # ... subtract remaining ALDs by adhoc allocation between spouses
        unit_total_alds = tax_unit("above_the_line_deductions", period)
        unit_remaining_alds = unit_total_alds - unit_total_personal_alds
        filing_status = person.tax_unit("filing_status", period)
        is_married = filing_status == filing_status.possible_values.JOINT
        is_head = person("is_tax_unit_head", period)
        is_spouse = person("is_tax_unit_spouse", period)
        allocated_alds = where(
            is_head | is_spouse,
            unit_remaining_alds / where(is_married, 2, 1),
            0,
        )
        fed_agi = gross_income - ind_total_personal_alds - allocated_alds
        # return MO AGI including MO additions and MO subtractions
        subtractions = person("mo_qualified_health_insurance_premiums", period)
        return max_(0, fed_agi - subtractions)



from policyengine_us.model_api import *


class total_itemized_taxable_income_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Total values of itemized taxable income deductions"
    unit = USD
    definition_period = YEAR

    adds = "gov.irs.deductions.itemized_deductions"



from policyengine_us.model_api import *


class ccdf_county_cluster(Variable):
    value_type = int
    entity = Household
    label = "County cluster for CCDF"
    definition_period = YEAR

    def formula(household, period, parameters):
        county = household("county", period).decode_to_str()
        cluster_mapping = parameters(period).gov.hhs.ccdf.county_cluster
        result = np.ones_like(county, dtype=int)
        valid_county = np.isin(
            county, np.array(list(cluster_mapping._children))
        )
        if valid_county.sum() > 0:
            result[valid_county] = cluster_mapping[county[valid_county]]
        return result



from policyengine_us.model_api import *


# Repealing Minnesota Bill HF1938 to pre 2023 rules
def create_mn_walz_hf1938_repeal() -> Reform:
    class mn_refundable_credits(Variable):
        value_type = float
        entity = TaxUnit
        label = "Minnesota refundable income tax credits"
        unit = USD
        definition_period = YEAR
        reference = (
            "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1ref_21.pdf"
            "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1ref_22.pdf"
        )
        defined_for = StateCode.MN

        def formula(tax_unit, period, parameters):
            if period.start.year >= 2023:
                instant_str = f"2022-01-01"
            else:
                instant_str = period
            p = parameters(instant_str).gov.states.mn.tax.income.credits
            return add(tax_unit, period, p.refundable)

    class mn_social_security_subtraction(Variable):
        value_type = float
        entity = TaxUnit
        label = "Minnesota social security subtraction"
        unit = USD
        definition_period = YEAR
        reference = (
            "https://www.taxformfinder.org/forms/2021/2021-minnesota-form-m1m.pdf"
            "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1m_22.pdf"
        )
        defined_for = StateCode.MN

        def formula(tax_unit, period, parameters):
            # specify parameters
            filing_status = tax_unit("filing_status", period)
            p = parameters(
                period
            ).gov.states.mn.tax.income.subtractions.social_security
            total_benefit_fraction = p.total_benefit_fraction
            income_amount = p.income_amount[filing_status]
            net_income_fraction = p.net_income_fraction
            alt_amount = p.alternative_amount[filing_status]
            # calculate subtraction amount (following "Worksheet for line 12")
            # ... US-taxable social security benefits
            us_taxable_oasdi = add(
                tax_unit, period, ["taxable_social_security"]
            )
            # ... alternative benefit subtraction amount
            us_gross_income = add(tax_unit, period, ["irs_gross_income"])
            adj_income = us_gross_income - us_taxable_oasdi
            total_oasdi = add(tax_unit, period, ["social_security"])
            oasdi_amount = total_oasdi * total_benefit_fraction
            tax_exempt_int = add(
                tax_unit, period, ["tax_exempt_interest_income"]
            )
            sum_income = adj_income + oasdi_amount + tax_exempt_int
            us_ald = tax_unit("above_the_line_deductions", period)
            student_loan_int = add(tax_unit, period, ["student_loan_interest"])
            mn_ald = max_(0, us_ald - student_loan_int)
            income = max_(0, sum_income - mn_ald)
            net_income = max_(0, income - income_amount)
            alt_sub_amt = max_(
                0, alt_amount - (net_income * net_income_fraction)
            )
            return min_(us_taxable_oasdi, alt_sub_amt)

    class mn_standard_deduction(Variable):
        value_type = float
        entity = TaxUnit
        label = "Minnesota standard deduction"
        unit = USD
        definition_period = YEAR
        reference = (
            "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1_inst_21.pdf"
            "https://www.revenue.state.mn.us/sites/default/files/2023-03/m1_inst_22.pdf"
        )
        defined_for = StateCode.MN

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.states.mn.tax.income.deductions.standard
            # ... calculate pre-limitation amount
            filing_status = tax_unit("filing_status", period)
            base_amt = p.base[filing_status]
            aged_blind_count = tax_unit("aged_blind_count", period)
            extra_amt = aged_blind_count * p.extra[filing_status]
            std_ded = base_amt + extra_amt
            # ... calculate standard deduction offset
            std_ded_offset = p.reduction.alternate.rate * std_ded
            agi = tax_unit("adjusted_gross_income", period)
            excess_agi = max_(
                0, agi - p.reduction.agi_threshold.low[filing_status]
            )
            excess_agi_offset = (
                p.reduction.excess_agi_fraction.low * excess_agi
            )
            offset = min_(std_ded_offset, excess_agi_offset)
            return max_(0, std_ded - offset)

    class mn_itemized_deductions(Variable):
        value_type = float
        entity = TaxUnit
        label = "Minnesota itemized deductions"
        unit = USD
        definition_period = YEAR
        reference = (
            "https://www.revenue.state.mn.us/sites/default/files/2021-12/m1_21_0.pdf"
            "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1_inst_21.pdf"
            "https://www.revenue.state.mn.us/sites/default/files/2022-12/m1_22.pdf"
            "https://www.revenue.state.mn.us/sites/default/files/2023-03/m1_inst_22.pdf"
        )
        defined_for = StateCode.MN

        def formula(tax_unit, period, parameters):
            # 2021 Form M1 instructions say:
            #   You may claim the Minnesota standard deduction or itemize
            #   your deductions on your Minnesota return. You will generally
            #   pay less Minnesota income tax if you take the larger of your
            #   itemized or standard deduction.
            # ... calculate pre-limitation itemized deductions
            itm_deds_less_salt = tax_unit(
                "itemized_deductions_less_salt", period
            )
            capped_property_taxes = tax_unit("capped_property_taxes", period)
            mn_itm_deds = itm_deds_less_salt + capped_property_taxes
            # ... calculate itemized deductions offset
            p = parameters(period).gov.states.mn.tax.income.deductions.itemized
            exempt_deds = add(
                tax_unit,
                period,
                ["medical_expense_deduction", "casualty_loss_deduction"],
            )
            net_deds = max_(0, mn_itm_deds - exempt_deds)
            net_deds_offset = p.reduction.alternate.rate * net_deds
            agi = tax_unit("adjusted_gross_income", period)
            filing_status = tax_unit("filing_status", period)
            excess_agi = max_(
                0, agi - p.reduction.agi_threshold.low[filing_status]
            )
            excess_agi_offset = (
                p.reduction.excess_agi_fraction.low * excess_agi
            )
            offset = min_(net_deds_offset, excess_agi_offset)
            return max_(0, mn_itm_deds - offset)

    class mn_cdcc(Variable):
        value_type = float
        entity = TaxUnit
        label = "Minnesota child and dependent care expense credit"
        unit = USD
        definition_period = YEAR
        reference = (
            "https://www.revenue.state.mn.us/sites/default/files/2023-02/m1cd_21.pdf"
            "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1cd_22_0.pdf"
        )
        defined_for = StateCode.MN

        def formula(tax_unit, period, parameters):
            p = parameters(period).gov.states.mn.tax.income.credits.cdcc
            person = tax_unit.members
            # determine eligibility for Minnesota CDCC
            filing_status = tax_unit("filing_status", period)
            eligible = filing_status != filing_status.possible_values.SEPARATE
            # calculate number of qualifying dependents
            # ... children
            age = person("age", period)
            qualifies_by_age = age < p.child_age
            # ... disability
            non_head = ~person("is_tax_unit_head", period)
            disabled = person("is_incapable_of_self_care", period)
            qualifies_by_disability = non_head & disabled
            dep_count = tax_unit.sum(
                qualifies_by_age | qualifies_by_disability
            )
            # calculate qualifying care expenses
            expense = tax_unit("tax_unit_childcare_expenses", period)
            # ... cap expense by number of qualifying dependents
            eligible_count = min_(dep_count, p.maximum_dependents)
            expense = min_(expense, p.maximum_expense * eligible_count)
            # ... cap expense by lower earnings of head and spouse if present
            expense = min_(expense, tax_unit("min_head_spouse_earned", period))
            # calculate pre-phaseout credit amount
            agi = tax_unit("adjusted_gross_income", period)
            pre_po_amount = expense * p.expense_fraction.calc(agi)
            # calculate post-phaseout credit amount
            excess_agi = max_(0, agi - p.phaseout_threshold)
            po_amount = excess_agi * p.phaseout_rate
            amount = max_(0, pre_po_amount - po_amount)
            # credit amount only for eligibles
            return eligible * amount

    class reform(Reform):
        def apply(self):
            self.neutralize_variable("mn_public_pension_subtraction")
            self.update_variable(mn_social_security_subtraction)
            self.update_variable(mn_refundable_credits)
            self.update_variable(mn_standard_deduction)
            self.update_variable(mn_itemized_deductions)
            self.update_variable(mn_cdcc)

    return reform


def create_mn_walz_hf1938_repeal_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_mn_walz_hf1938_repeal()

    p = parameters(period).gov.contrib.states.mn.walz.hf1938

    if p.repeal:
        return create_mn_walz_hf1938_repeal()
    else:
        return None


mn_walz_hf1938 = create_mn_walz_hf1938_repeal_reform(None, None, bypass=True)



from policyengine_us.model_api import *


class self_employed_pension_contributions(Variable):
    value_type = float
    entity = Person
    label = "Self-employed pension contributions"
    unit = USD
    documentation = "Pension plan contributions associated with plans for the self employed."
    definition_period = YEAR



from policyengine_us.model_api import *


class wa_capital_gains_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Washington capital gains tax"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.WA

    def formula(tax_unit, period, parameters):
        in_effect = parameters(period).gov.states.wa.tax.income.in_effect
        if in_effect:
            p = parameters(period).gov.states.wa.tax.income.capital_gains
            ltcg = add(tax_unit, period, ["long_term_capital_gains"])
            # Deduct charitable contributions.
            charitable_contributions = add(
                tax_unit,
                period,
                ["charitable_cash_donations", "charitable_non_cash_donations"],
            )
            charitable_contributions_above_exemption = max_(
                0, charitable_contributions - p.deductions.charitable.exemption
            )
            charitable_deduction = min_(
                charitable_contributions_above_exemption,
                p.deductions.charitable.cap,
            )
            # Get base after standard and charitable deductions.
            taxable_ltcg = max_(
                0, ltcg - charitable_deduction - p.deductions.standard
            )
            return taxable_ltcg * p.rate
        return 0



from policyengine_us.model_api import *


class pr_mortgage_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Puerto Rico home mortgage deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://law.justia.com/codes/puerto-rico/title-thirteen/subtitle-17/part-ii/chapter-1005/subchapter-c/30135/"  # (1)(C)
    defined_for = StateCode.PR

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.territories.pr.tax.income.taxable_income.deductions.mortgage

        agi = tax_unit("pr_agi", period)
        limit = agi * p.floor  # plus any income excluded from AGI

        mortgage_interest = add(
            tax_unit, period, ["deductible_mortgage_interest"]
        )
        person = tax_unit.members
        age = person("age", period)

        is_senior_exception = tax_unit.any(
            person("is_tax_unit_head_or_spouse", period)
            & (age >= p.age_threshold)
        )
        # True if any member fulfills condition

        # people 65+ years old are capped at max deduction amount
        interest_capped_by_max = min_(mortgage_interest, p.max)
        # otherwise capped by both AGI and max deduction amount
        combined_cap = min_(limit, p.max)
        interest_capped_by_combined = min_(mortgage_interest, combined_cap)
        return where(
            is_senior_exception,
            interest_capped_by_max,
            interest_capped_by_combined,
        )



from policyengine_us.model_api import *


class ga_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Georgia refundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.GA



from policyengine_us.model_api import *


class oh_has_taken_oh_lump_sum_credits(Variable):
    value_type = bool
    entity = Person
    label = "Whether a person has taken Ohio lump sum credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2021/pit-it1040-booklet.pdf#page=20",
        "https://codes.ohio.gov/ohio-revised-code/section-5747.055",
    )
    defined_for = StateCode.OH



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_
from policyengine_core.periods import instant
import numpy as np


def create_reconciled_ssn_for_llc_and_aoc() -> Reform:
    # Amend the variables themselves, as the *_eligible flags indicate
    # the ability of individuals to count their educational expenses.
    # These reforms set identification requirements at the filer level.
    class lifetime_learning_credit(Variable):
        value_type = float
        entity = TaxUnit
        label = "Lifetime Learning Credit"
        unit = USD
        documentation = "Value of the non-refundable Lifetime Learning Credit"
        definition_period = YEAR
        reference = "https://www.law.cornell.edu/uscode/text/26/25A#c"
        defined_for = "filer_meets_llc_and_aoc_identification_requirements"

        def formula(tax_unit, period, parameters):
            education = parameters(period).gov.irs.credits.education
            llc = education.lifetime_learning_credit
            person = tax_unit.members
            is_aoc_eligible = person(
                "is_eligible_for_american_opportunity_credit", period
            )
            eligible_expenses = tax_unit.sum(
                person("qualified_tuition_expenses", period) * ~is_aoc_eligible
            )
            capped_expenses = min_(llc.expense_limit, eligible_expenses)
            maximum_amount = llc.rate * capped_expenses
            phase_out = tax_unit("education_credit_phase_out", period)
            if llc.abolition:
                return 0
            return max_(0, maximum_amount * (1 - phase_out))

    class american_opportunity_credit(Variable):
        value_type = float
        entity = TaxUnit
        label = "American Opportunity Credit"
        unit = USD
        documentation = "Total value of the American Opportunity Credit"
        definition_period = YEAR
        reference = "https://www.law.cornell.edu/uscode/text/26/25A#b"
        defined_for = "filer_meets_llc_and_aoc_identification_requirements"

        def formula(tax_unit, period, parameters):
            education = parameters(period).gov.irs.credits.education
            aoc = education.american_opportunity_credit
            person = tax_unit.members
            is_eligible = person(
                "is_eligible_for_american_opportunity_credit", period
            )
            tuition_expenses = (
                person("qualified_tuition_expenses", period) * is_eligible
            )
            maximum_amount_per_student = aoc.amount.calc(tuition_expenses)
            maximum_amount = tax_unit.sum(maximum_amount_per_student)
            phase_out = tax_unit("education_credit_phase_out", period)
            if aoc.abolition:
                return 0
            return max_(0, maximum_amount * (1 - phase_out))

    class filer_meets_llc_and_aoc_identification_requirements(Variable):
        value_type = bool
        entity = TaxUnit
        definition_period = YEAR
        label = "Filer meets LLC and AOC identification requirements"

        def formula(tax_unit, period, parameters):
            # Both head and spouse in the tax unit must have valid SSN card type to be eligible for the CTC
            person = tax_unit.members
            is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
            eligible_ssn_card_type = person(
                "meets_llc_and_aoc_identification_requirements", period
            )
            ineligible_head_or_spouse = (
                is_head_or_spouse & ~eligible_ssn_card_type
            )
            return tax_unit.sum(ineligible_head_or_spouse) == 0

    class meets_llc_and_aoc_identification_requirements(Variable):
        value_type = bool
        entity = Person
        definition_period = YEAR
        label = "Person meets LLC and AOC identification requirements"
        reference = "https://docs.house.gov/meetings/WM/WM00/20250513/118260/BILLS-119CommitteePrintih.pdf#page=4"

        def formula(person, period, parameters):
            ssn_card_type = person("ssn_card_type", period)
            ssn_card_types = ssn_card_type.possible_values
            citizen = ssn_card_type == ssn_card_types.CITIZEN
            non_citizen_valid_ead = (
                ssn_card_type == ssn_card_types.NON_CITIZEN_VALID_EAD
            )
            return citizen | non_citizen_valid_ead

    class reform(Reform):
        def apply(self):
            self.update_variable(lifetime_learning_credit)
            self.update_variable(american_opportunity_credit)
            self.update_variable(
                filer_meets_llc_and_aoc_identification_requirements
            )
            self.update_variable(meets_llc_and_aoc_identification_requirements)

    return reform


def create_reconciled_ssn_for_llc_and_aoc_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_reconciled_ssn_for_llc_and_aoc()

    p = parameters.gov.contrib.reconciliation.ssn_for_llc_and_aoc

    reform_active = False
    current_period = period_(period)

    for i in range(5):
        if p(current_period).in_effect:
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    if reform_active:
        return create_reconciled_ssn_for_llc_and_aoc()
    else:
        return None


reconciled_ssn_for_llc_and_aoc = create_reconciled_ssn_for_llc_and_aoc_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class gi_cash_assistance(Variable):
    value_type = float
    entity = Person
    label = "guaranteed income / cash assistance income"
    unit = USD
    documentation = "Income from guaranteed income or cash assistance pilots"
    definition_period = YEAR



from policyengine_us.model_api import *


class state_sales_tax_income_bracket(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    label = "State Sales Tax Income Bracket"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.irs.deductions.itemized.salt_and_real_estate.state_sales_tax_table
        income = add(tax_unit, period, p.income_sources)
        return p.income_bracket.calc(income)



from policyengine_us.model_api import *


class mn_public_pension_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minnesota public pension subtraction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.taxformfinder.org/forms/2021/2021-minnesota-form-m1m.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1m_22.pdf"
    )
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        taxable_public_pension_income = add(
            tax_unit, period, ["taxable_public_pension_income"]
        )
        p = parameters(
            period
        ).gov.states.mn.tax.income.subtractions.pension_income
        filing_status = tax_unit("filing_status", period)
        cap = p.cap[filing_status]
        capped_pension_income = min_(taxable_public_pension_income, cap)
        agi = tax_unit("adjusted_gross_income", period)
        agi_threshold = p.reduction.start[filing_status]
        excess = max_(0, agi - agi_threshold)
        reduced_agi_fraction = np.ceil(excess / p.reduction.increment)
        percentage = min_(p.reduction.rate * reduced_agi_fraction, 1)
        reduction = capped_pension_income * percentage
        return max_(capped_pension_income - reduction, 0)



from policyengine_us.model_api import *


class spouse_separate_tax_unit_size(Variable):
    value_type = int
    entity = TaxUnit
    definition_period = YEAR
    label = "Size of spouse's tax unit if they file separately"

    # Assume the same characteristics.
    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        separate = filing_status == filing_status.possible_values.SEPARATE
        return tax_unit("tax_unit_size", period) * separate



from policyengine_us.model_api import *


class overtime_income_deduction_ssn_requirement_met(Variable):
    value_type = bool
    entity = TaxUnit
    label = "SSN requirement met for the overtime income deduction"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        ssn_card_type = person("ssn_card_type", period)
        ssn_card_str = ssn_card_type.decode_to_str()
        p = parameters(period).gov.irs.deductions.overtime_income
        eligible_ssn_card_type = np.isin(
            ssn_card_str, p.eligible_ssn_card_type
        )
        joint = tax_unit("tax_unit_is_joint", period)
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        eligible_head_or_spouse = eligible_ssn_card_type & head_or_spouse
        return where(
            joint,
            tax_unit.sum(eligible_head_or_spouse) == 2,
            tax_unit.any(eligible_head_or_spouse),
        )



from policyengine_us.model_api import *


class mo_ptc_net_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Missouri property tax credit net income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.mo.gov/forms/MO-PTS_2021.pdf",
        "https://dor.mo.gov/forms/4711_2021.pdf",
        "https://revisor.mo.gov/main/OneSection.aspx?section=135.010&bid=6435&hl=property+tax+credit%u2044",
    )
    defined_for = StateCode.MO

    def formula(tax_unit, period, parameters):
        gross_income = tax_unit("mo_ptc_gross_income", period)
        income_offset = tax_unit("mo_ptc_income_offset", period)
        return max_(0, gross_income - income_offset)  # Form MO-PTS, line 10



from policyengine_us.model_api import *


class ky_standard_deduction_indiv(Variable):
    value_type = float
    entity = Person
    label = "Kentucky standard deduction when married couples file separately"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.KY

    def formula(person, period, parameters):
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        p = parameters(period).gov.states.ky.tax.income.deductions
        return head_or_spouse * p.standard



from policyengine_us.model_api import *


class il_k12_education_expense_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Illinois K-12 Education Expense Credit"
    unit = USD
    definition_period = YEAR

    defined_for = "il_is_exemption_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.il.tax.income.credits.k12
        tuition_and_fees = tax_unit("k12_tuition_and_fees", period)
        reduced_tuition_and_fees = max_(0, tuition_and_fees - p.reduction)
        k12_credit = min_(reduced_tuition_and_fees * p.rate, p.cap)
        pre_credit_tax = tax_unit(
            "il_income_tax_before_non_refundable_credits", period
        )
        il_property_tax_credit = tax_unit("il_property_tax_credit", period)
        avail_tax = max_(0, pre_credit_tax - il_property_tax_credit)
        return min_(avail_tax, k12_credit)



from policyengine_us.model_api import *


class employment_income_behavioral_response(Variable):
    value_type = float
    entity = Person
    label = "employment income behavioral response"
    unit = USD
    definition_period = YEAR

    def formula(person, period, parameters):
        lsr = person("labor_supply_behavioral_response", period)
        raw_earnings = add(
            person,
            period,
            [
                "employment_income_before_lsr",
                "self_employment_income_before_lsr",
            ],
        )
        earnings = max_(raw_earnings, 0)
        employment_income = person("employment_income_before_lsr", period)
        emp_share = np.ones_like(earnings)
        mask = earnings > 0
        emp_share[mask] = employment_income[mask] / earnings[mask]
        return lsr * emp_share



from policyengine_us.model_api import *


class own_children_in_household(Variable):
    value_type = int
    entity = Person
    label = "Count of one's own children in the household"
    definition_period = YEAR



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_
from policyengine_core.periods import instant


def create_fisc_act() -> Reform:
    class family_income_supplement_credit_base_amount(Variable):
        value_type = float
        entity = Person
        label = "FISC Act family income supplement base amount"
        unit = USD
        definition_period = YEAR
        reference = "https://golden.house.gov/sites/evo-subsites/golden.house.gov/files/evo-media-document/GoldenFISC.pdf"

        def formula(person, period, parameters):
            # Calculate the base amount
            is_dependent = person("is_tax_unit_dependent", period)
            age = person("age", period)
            p = parameters(
                period
            ).gov.contrib.congress.golden.fisc_act.family_income_supplement
            eligible_dependent = (age < p.child_age_limit) & is_dependent
            return p.amount.base.calc(age) * eligible_dependent

    class family_income_supplement_credit(Variable):
        value_type = float
        entity = TaxUnit
        label = "FISC Act family income supplement"
        unit = USD
        definition_period = YEAR
        reference = "https://golden.house.gov/sites/evo-subsites/golden.house.gov/files/evo-media-document/GoldenFISC.pdf"

        def formula(tax_unit, period, parameters):

            p = parameters(
                period
            ).gov.contrib.congress.golden.fisc_act.family_income_supplement
            # A joint bonus is applied to the base amount
            filing_status = tax_unit("filing_status", period)
            joint = filing_status == filing_status.possible_values.JOINT
            base_amount = add(
                tax_unit,
                period,
                [
                    "family_income_supplement_credit_base_amount",
                ],
            )
            base_amount_with_marriage_bonus = base_amount * (
                1 + joint * p.marriage_bonus_rate
            )
            agi = tax_unit("adjusted_gross_income", period)
            phase_out = where(
                joint, p.phase_out.joint.calc(agi), p.phase_out.other.calc(agi)
            )
            phased_out_amount = max_(
                base_amount_with_marriage_bonus - phase_out, 0
            )
            agi_threshold = max_(0, p.agi_fraction * agi)
            capped_credit = min_(phased_out_amount, agi_threshold)
            # The credit is limited to eligible caregivers
            person = tax_unit.members
            age = person("age", period)
            age_eligible = age >= p.caregiver_age_threshold
            head_or_spouse = person("is_tax_unit_head_or_spouse", period)
            eligible_caregiver_present = tax_unit.any(
                age_eligible & head_or_spouse
            )
            return capped_credit * eligible_caregiver_present

    def modify_parameters(parameters):
        parameters.gov.irs.credits.refundable.update(
            start=instant("2026-01-01"),
            stop=instant("2035-12-31"),
            value=[
                "eitc",
                "refundable_american_opportunity_credit",
                "recovery_rebate_credit",
                "refundable_payroll_tax_credit",
                "family_income_supplement_credit",
            ],
        )
        return parameters

    class reform(Reform):
        def apply(self):
            self.update_variable(family_income_supplement_credit_base_amount)
            self.update_variable(family_income_supplement_credit)
            self.modify_parameters(modify_parameters)
            self.neutralize_variable("refundable_ctc")
            self.neutralize_variable("non_refundable_ctc")

    return reform


def create_fisc_act_reform(parameters, period, bypass: bool = False):
    if bypass:
        return create_fisc_act()

    p = parameters.gov.contrib.congress.golden.fisc_act

    reform_active = False
    current_period = period_(period)

    for i in range(5):
        if p(current_period).in_effect:
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    if reform_active:
        return create_fisc_act()
    else:
        return None


fisc_act = create_fisc_act_reform(None, None, bypass=True)



from policyengine_us.model_api import *


class pr_gross_income_person(Variable):
    value_type = float
    entity = Person
    label = "Puerto Rico gross income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.PR
    reference = "https://law.justia.com/codes/puerto-rico/title-thirteen/subtitle-17/part-ii/chapter-1005/subchapter-a/30101/"

    adds = "gov.territories.pr.tax.income.gross_income.sources"



from policyengine_us.model_api import *


class slcsp_rating_area_la_county(Variable):
    value_type = int
    entity = Household
    label = (
        "Second-lowest ACA silver-plan cost rating area in Los Angeles County"
    )
    definition_period = YEAR
    defined_for = "in_la"

    def formula(household, period, parameters):
        zip3 = household("three_digit_zip_code", period)
        p = parameters(period).gov.aca
        is_in_la = household("in_la", period)
        result = np.empty(zip3.shape, dtype=int)
        result[is_in_la] = p.la_county_rating_area[zip3[is_in_la]]
        result[~is_in_la] = 0
        return result



from policyengine_us.model_api import *


class household_count(Variable):
    value_type = float
    entity = Household
    label = "Households represented"
    definition_period = YEAR
    default_value = 1.0



from policyengine_us.model_api import *


class limited_capital_loss(Variable):
    value_type = float
    entity = TaxUnit
    label = "Limited capital loss deduction"
    unit = USD
    documentation = "The capital loss deductible from gross income."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/1211"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs
        filing_status = tax_unit("filing_status", period)
        max_loss = p.ald.loss.capital.max[filing_status]
        return min_(max_loss, add(tax_unit, period, ["capital_losses"]))



from policyengine_us.model_api import *


class ok_income_tax_before_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Oklahoma income tax before refundable credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/past-year/2021/511-Pkt-2021.pdf"
        "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-Pkt.pdf"
    )
    defined_for = StateCode.OK

    def formula(tax_unit, period, parameters):
        itax_before_credits = tax_unit("ok_income_tax_before_credits", period)
        nonrefundable_credits = tax_unit("ok_nonrefundable_credits", period)
        return max_(0, itax_before_credits - nonrefundable_credits)



from policyengine_us.model_api import *


class spm_unit_benefits(Variable):
    value_type = float
    entity = SPMUnit
    label = "Benefits"
    definition_period = YEAR
    unit = USD

    def formula(spm_unit, period, parameters):
        BENEFITS = [
            "social_security",
            "ssi",
            "ma_state_supplement",  # Massachusetts benefits
            # California programs.
            "ca_cvrp",  # California Clean Vehicle Rebate Project.
            # Colorado programs.
            "co_ccap_subsidy",
            "co_state_supplement",
            "co_oap",
            "snap",
            "wic",
            "free_school_meals",
            "reduced_price_school_meals",
            "spm_unit_broadband_subsidy",
            "spm_unit_energy_subsidy",
            "tanf",
            "high_efficiency_electric_home_rebate",
            "residential_efficiency_electrification_rebate",
            "unemployment_compensation",
            # One-time energy relief payments.
            # Paid at the same time as the Alaska Permanent Fund Dividend,
            # which is part of IRS gross income.
            "ak_energy_relief",
            # Contributed.
            "basic_income",
            "ny_drive_clean_rebate",
        ]
        if parameters(period).gov.contrib.ubi_center.flat_tax.deduct_ptc:
            BENEFITS.append("aca_ptc")
        if not parameters(period).gov.hud.abolition:
            BENEFITS.append("spm_unit_capped_housing_subsidy")
        return add(spm_unit, period, BENEFITS)



from policyengine_us.model_api import *


class ut_personal_exemption_additional_dependent_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Utah additional dependent personal exemption eligible"
    defined_for = StateCode.UT
    definition_period = YEAR
    reference = "https://le.utah.gov/xcode/Title59/Chapter10/59-10-S1018.html"  # 59-10-1018 (1)(g)

    def formula(person, period, parameters):
        is_dependent = person("is_tax_unit_dependent", period)
        birth_year = person("birth_year", period)
        born_this_year = birth_year == period.start.year
        return is_dependent & born_this_year



from policyengine_us.model_api import *


class qualified_adoption_assistance_expense(Variable):
    value_type = float
    entity = Person
    label = "Qualified adoption expense"
    unit = USD
    definition_period = YEAR
    documentation = "Qualified adoption expense (as defined in 26 U.S. Code § 23(d)) made pursuant to an adoption assistance program."
    reference = "https://www.law.cornell.edu/uscode/text/26/23#d"

    # Note that this is an above-the-line deduction. There is more
    # generous tax relief for qualified adoption expenses involving
    # children with special needs, which is not modelled here.



from policyengine_us.model_api import *


class va_income_tax_before_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia income tax before credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://law.lis.virginia.gov/vacodefull/title58.1/chapter3/article2/"
    )
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        tax_before_non_refundable_credits = tax_unit(
            "va_income_tax_before_non_refundable_credits", period
        )
        non_refundable_credits = tax_unit("va_non_refundable_credits", period)
        spouse_tax_adjustment = tax_unit("va_spouse_tax_adjustment", period)
        return max_(
            tax_before_non_refundable_credits
            - non_refundable_credits
            - spouse_tax_adjustment,
            0,
        )



from policyengine_us.model_api import *


class wv_senior_citizen_disability_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "West Virginia senior citizen or disability deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.WV

    adds = ["wv_senior_citizen_disability_deduction_person"]



from policyengine_us.model_api import *


class takes_up_dc_ptc(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Takes up the DC property tax credit"
    definition_period = YEAR
    defined_for = StateCode.DC
    default_value = True



from policyengine_us.model_api import *


class dc_ccsp_asset_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for DC Child Care Subsidy Program (CCSP) due to income"
    definition_period = MONTH
    reference = "https://osse.dc.gov/sites/default/files/dc/sites/osse/publication/attachments/DC%20Child%20Care%20Subsidy%20Program%20Policy%20Manual.pdf#page=12"
    defined_for = StateCode.DC

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.dc.dhs.ccsp.assets
        asset = spm_unit("dc_ccsp_assets", period)
        return asset <= p.limit



from policyengine_us.model_api import *


class taxable_403b_distributions(Variable):
    value_type = float
    entity = Person
    label = "Taxable 403(b) distributions"
    unit = USD
    documentation = (
        "Taxable distributions from 403b accounts (typically traditional)."
    )
    definition_period = YEAR



from policyengine_us.model_api import *


class fdpir(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    label = "Food Distribution Program on Indian Reservations"
    unit = USD
    documentation = (
        "Benefit value of the Food Distribution Program on Indian Reservations"
    )



from policyengine_us.model_api import *


class wi_earned_income_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Wisconsin earned income credit (WI EITC)"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.wi.gov/TaxForms2021/2021-Form1f.pdf#page=2"
        "https://www.revenue.wi.gov/TaxForms2021/2021-Form1-Inst.pdf#page=26"
        "https://www.revenue.wi.gov/TaxForms2022/2022-Form1f.pdf#page=2"
        "https://www.revenue.wi.gov/TaxForms2022/2022-Form1-Inst.pdf#page=26"
        "https://docs.legis.wisconsin.gov/misc/lfb/informational_papers/january_2023/0002_individual_income_tax_informational_paper_2.pdf"
    )
    defined_for = StateCode.WI

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.wi.tax.income.credits.earned_income
        inv_income = tax_unit("eitc_relevant_investment_income", period)
        # In 2023 Wisconsin adopted the federal EITC investment income limit
        if p.apply_federal_investment_income_limit:
            p_irs = parameters(period).gov.irs.credits.eitc.phase_out
            threshold = p_irs.max_investment_income
        else:
            threshold = p.investment_income_limit
        eligible = inv_income <= threshold
        federal_eitc = tax_unit("eitc", period)
        child_count = tax_unit("eitc_child_count", period)
        wi_frac = p.fraction.calc(child_count)
        return eligible * federal_eitc * wi_frac



from policyengine_us.model_api import *


class ct_social_security_benefit_adjustment(Variable):
    value_type = float
    entity = TaxUnit
    unit = USD
    label = "Connecticut social security benefit adjustment"
    reference = (
        # Connecticut General Statutes, Chapter 229, Sec. 12-701, (20), (b), (x), (iii) and (iv)
        "https://www.cga.ct.gov/current/pub/chap_229.htm#sec_12-701"
        # 2022 Form CT-1040 Connecticut Resident Income Tax Return Instructions
        "https://portal.ct.gov/-/media/DRS/Forms/2022/Income/2022-CT-1040-Instructions_1222.pdf#page=24"
    )
    definition_period = YEAR
    defined_for = StateCode.CT

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        # Part A
        social_security = tax_unit("tax_unit_social_security", period)
        # Part B
        p_irs = parameters(period).gov.irs.social_security.taxability
        ss_combined_income_excess = tax_unit(
            "tax_unit_ss_combined_income_excess", period
        )
        # Part C
        capped_social_security = min_(
            social_security, ss_combined_income_excess
        )
        # Part D
        p = parameters(
            period
        ).gov.states.ct.tax.income.subtractions.social_security
        capped_social_security_portion = capped_social_security * p.rate
        # Part E (Line 18 from federal social security worksheet)
        gross_ss = tax_unit("tax_unit_social_security", period)
        adjusted_gross_social_security = gross_ss * p_irs.rate.additional
        # Part F
        reduced_taxable_social_security = max_(
            adjusted_gross_social_security - capped_social_security_portion,
            0,
        )
        # Filers with AGI below the threshold can subtract the full amount of their
        # taxable social security benefits
        agi = tax_unit("adjusted_gross_income", period)
        full_adjustment_eligible = agi < p.reduction_threshold[filing_status]
        taxable_social_security = tax_unit(
            "tax_unit_taxable_social_security", period
        )
        return where(
            full_adjustment_eligible,
            taxable_social_security,
            reduced_taxable_social_security,
        )



from policyengine_us.model_api import *


class aca_magi(Variable):
    value_type = float
    entity = TaxUnit
    label = "ACA-related modified AGI for this tax unit"
    definition_period = YEAR
    adds = ["medicaid_magi"]



from policyengine_us.model_api import *


class id_grocery_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Idaho grocery credit"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.ID
    reference = (
        "https://law.justia.com/codes/idaho/2022/title-63/chapter-30/section-63-3024a/",
        "https://tax.idaho.gov/wp-content/uploads/forms/EFO00089/EFO00089_12-30-2022.pdf#page=7",
    )

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        qualified_months = person("id_grocery_credit_qualified_months", period)
        p = parameters(period).gov.states.id.tax.income.credits.grocery.aged
        if p.in_effect:
            full_amount = add(
                person,
                period,
                ["id_grocery_credit_base", "id_grocery_credit_aged"],
            )
        else:
            full_amount = person("id_grocery_credit_base", period)
        credit_value = full_amount * (qualified_months / MONTHS_IN_YEAR)
        return tax_unit.sum(credit_value)



from policyengine_us.model_api import *


class safmr_used_for_hcv(Variable):
    value_type = bool
    entity = Household
    label = "Small area fair market rent used for purposes of the Housing Choice Voucher Program"
    definition_period = YEAR



from policyengine_us.model_api import *


class ny_itemized_deductions_reduction_applies(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Whether the reduction to the New York itemized deductions applies"
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/615"  # (f)
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.ny.tax.income.deductions.itemized.reduction.incremental.lower
        agi = tax_unit("ny_agi", period)
        filing_status = tax_unit("filing_status", period)
        lower_income_threshold = p.income_threshold[filing_status]
        # No reduction applies if the AGI is less than the first reduction threshold
        return agi > lower_income_threshold



from policyengine_us.model_api import *


class dc_income_tax_before_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "DC income tax before credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/52926_D-40_12.21.21_Final_Rev011122.pdf#page=36"
        "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/2022_D-40_Booklet_Final_blk_01_23_23_Ordc.pdf#page=34"
    )
    defined_for = StateCode.DC

    def formula(tax_unit, period, parameters):
        return where(
            tax_unit("dc_files_separately", period),
            tax_unit("dc_income_tax_before_credits_indiv", period),
            tax_unit("dc_income_tax_before_credits_joint", period),
        )



from policyengine_us.model_api import *


class ca_calworks_child_care_age_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "California CalWORKs Child Care SPMUnit Age Eligibility"
    definition_period = YEAR
    defined_for = StateCode.CA
    reference = "http://epolicy.dpss.lacounty.gov/epolicy/epolicy/server/general/projects_responsive/ePolicyMaster/index.htm?&area=general&type=responsivehelp&ctxid=&project=ePolicyMaster#t=mergedProjects%2FChild%20Care%2FChild_Care%2F1210_Overview%2F1210_Overview.htm%23Backgroundbc-3&rhtocid=_3_3_0_2"

    def formula(spm_unit, period, parameters):
        person = spm_unit.members
        child_age_eligible = person(
            "ca_calworks_child_care_child_age_eligible", period
        )

        return spm_unit.any(child_age_eligible)



import pandas as pd
import requests
from tqdm import tqdm
from pathlib import Path
import os
import zipfile
import shutil

STATE_NAMES = [
    "Alabama",
    "Alaska",
    "Arizona",
    "Arkansas",
    "California",
    "Colorado",
    "Connecticut",
    "Delaware",
    "Florida",
    "Georgia",
    "Hawaii",
    "Idaho",
    "Illinois",
    "Indiana",
    "Iowa",
    "Kansas",
    "Kentucky",
    "Louisiana",
    "Maine",
    "Maryland",
    "Massachusetts",
    "Michigan",
    "Minnesota",
    "Mississippi",
    "Missouri",
    "Montana",
    "Nebraska",
    "Nevada",
    "New Hampshire",
    "New Jersey",
    "New Mexico",
    "New York",
    "North Carolina",
    "North Dakota",
    "Ohio",
    "Oklahoma",
    "Oregon",
    "Pennsylvania",
    "Rhode Island",
    "South Carolina",
    "South Dakota",
    "Tennessee",
    "Texas",
    "Utah",
    "Vermont",
    "Virginia",
    "Washington",
    "West Virginia",
    "Wisconsin",
    "Wyoming",
]
STATE_CODES = [
    "AL",
    "AK",
    "AZ",
    "AR",
    "CA",
    "CO",
    "CT",
    "DE",
    "FL",
    "GA",
    "HI",
    "ID",
    "IL",
    "IN",
    "IA",
    "KS",
    "KY",
    "LA",
    "ME",
    "MD",
    "MA",
    "MI",
    "MN",
    "MS",
    "MO",
    "MT",
    "NE",
    "NV",
    "NH",
    "NJ",
    "NM",
    "NY",
    "NC",
    "ND",
    "OH",
    "OK",
    "OR",
    "PA",
    "RI",
    "SC",
    "SD",
    "TN",
    "TX",
    "UT",
    "VT",
    "VA",
    "WA",
    "WV",
    "WI",
    "WY",
]
STATE_CODES = [x.lower() for x in STATE_CODES]
DATA_FOLDER = Path("data")
DATA_FOLDER.mkdir(exist_ok=True)

dfs = []
for state_name, state_code in tqdm(
    zip(STATE_NAMES, STATE_CODES), desc="Downloading Census data"
):
    data_url = f"https://www2.census.gov/programs-surveys/decennial/2020/data/01-Redistricting_File--PL_94-171/{state_name}/{state_code}2020.pl.zip"
    # Download the file and save to a folder called "block_level_population_data_by_state/"
    r = requests.get(data_url)
    with open(DATA_FOLDER / f"{state_code}2020.pl.zip", "wb") as f:
        f.write(r.content)
    # Unzip the file
    with zipfile.ZipFile(
        DATA_FOLDER / f"{state_code}2020.pl.zip", "r"
    ) as zip_ref:
        zip_ref.extractall(DATA_FOLDER / f"{state_code}2020.pl")
    # Delete the zip file
    os.remove(DATA_FOLDER / f"{state_code}2020.pl.zip")
    # Read the file
    df = pd.read_csv(
        DATA_FOLDER / f"{state_code}2020.pl/{state_code}geo2020.pl",
        sep="|",
        low_memory=False,
        encoding="ISO-8859-1",
    )
    df["state"] = state_code
    dfs += [df]
    full_df = pd.concat(dfs)
    full_df.to_csv(DATA_FOLDER / "50_state_block_data.csv", index=False)
    shutil.rmtree(DATA_FOLDER / f"{state_code}2020.pl")



from policyengine_us.model_api import *


class pr_medical_expense_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Puerto Rico medical expense deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://law.justia.com/codes/puerto-rico/title-thirteen/subtitle-17/part-ii/chapter-1005/subchapter-c/30135/"  # (4)
    defined_for = StateCode.PR

    def formula(tax_unit, period, parameters):
        expense = add(tax_unit, period, ["medical_out_of_pocket_expenses"])
        p = parameters(
            period
        ).gov.territories.pr.tax.income.taxable_income.deductions.medical
        medical_floor = p.floor * tax_unit("pr_agi", period)
        return max_(0, expense - medical_floor)



from policyengine_us.model_api import *


class la_general_relief_housing_subsidy_base_amount_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for the Los Angeles County General Relief Housing Subsidy based on the base amount requirements"
    definition_period = MONTH
    # Person has to be a resident of LA County
    defined_for = "la_general_relief_eligible"
    reference = "https://dpss.lacounty.gov/en/cash/gr/housing.html"

    def formula(spm_unit, period, parameters):
        # General relief cannot be under the rent subsidy amounts
        gr_base_amount = spm_unit("la_general_relief_base_amount", period)
        p = parameters(
            period
        ).gov.local.ca.la.general_relief.housing_subsidy.rent_contribution
        married = add(spm_unit, period, ["is_married"])
        rent_contribution = where(married, p.married, p.single)
        return gr_base_amount >= rent_contribution



from policyengine_us.model_api import *


class new_clean_vehicle_credit_potential(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Potential value of the New clean vehicle credit"
    documentation = (
        "Nonrefundable credit for the purchase of a new clean vehicle"
    )
    unit = USD
    reference = (
        "https://www.law.cornell.edu/uscode/text/26/30D",
        "https://www.democrats.senate.gov/imo/media/doc/inflation_reduction_act_of_2022.pdf#page=373",
    )
    defined_for = "new_clean_vehicle_credit_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.credits.clean_vehicle.new
        battery_component_percent = tax_unit(
            "new_clean_vehicle_battery_components_made_in_north_america",
            period,
        )
        meets_battery_component_test = (
            battery_component_percent >= p.battery_components.threshold
        )
        critical_minerals_percent = tax_unit(
            "new_clean_vehicle_battery_critical_minerals_extracted_in_trading_partner_country",
            period,
        )
        meets_critical_minerals_test = (
            critical_minerals_percent >= p.critical_minerals.threshold
        )
        battery_components_credit = (
            meets_battery_component_test * p.battery_components.amount
        )
        critical_minerals_credit = (
            meets_critical_minerals_test * p.critical_minerals.amount
        )
        # PRE-IRA POLICY:
        # Amount per kWh of EV battery capacity.
        # "In the case of a vehicle which draws propulsion energy from a
        # battery with not less than 5 kilowatt hours of capacity, the amount
        # determined under this paragraph is $417, plus $417 for each kilowatt
        # hour of capacity in excess of 5 kilowatt hours."
        capacity = tax_unit("new_clean_vehicle_battery_capacity", period)
        kwh_excess = max_(capacity - p.capacity_bonus.kwh_threshold + 1, 0)
        kwh_bonus = min_(
            p.capacity_bonus.amount * np.floor(kwh_excess),
            p.capacity_bonus.max,
        )
        # Sum all elements.
        return (
            p.base_amount
            + kwh_bonus
            + battery_components_credit
            + critical_minerals_credit
        )
